import io

import RobotInfo
from RobotLogger import robot_logger as logger
from Environment.AlphAI import Communication
import gpiozero
from LedStripe import colors_by_name
from ThreadUnit import ThreadUnit
from collections import deque
from picamera import PiCamera
import RPi.GPIO as GPIO
import numpy as np
import subprocess
import threading
import time

try:
    from RobotBuzzer import Buzzer
except ImportError:
    logger.warning("Tonal buzzer not ready in this release yet")


    class Buzzer:
        def __init__(self, pin):
            GPIO.setup(pin, GPIO.OUT)

try:
    from Bluetooth import Bluetooth, BluetoothInitError
    import Bluetooth_Scan

    _bluetooth = Bluetooth()
    bluetooth_ok = True
except:
    bluetooth_ok = False

try:
    from LedStripe import LedStripe

    led_stripe_ok = True
except ImportError as err:
    LedStripe = None
    led_stripe_ok = False

try:
    import UWB

    uwb_ok = True
    accelerometer_ok = True
except ImportError as err:
    UWB = None
    uwb_ok = False
    accelerometer_ok = False

_speed_buffer_size = 5
_speed_integrate_time = .4

COLOR = "color"
LEFT_RIGHT_BIAS = "left_right_bias"

# alias for IRport flag class
IRP = Communication.TracerPort


# noinspection PyPep8Naming
class AlphaBot:

    # Constructor
    def __init__(self, led_stripe=None):
        logger.info('Init AlphaBot')

        self.hardware = None
        self._read_hardware()
        self.BUZ = 4

        if self.hardware == "AlphaBot2":
            self.has_speed = False
            self.has_ir_line = True
            self.has_ir_front = False
            self.has_led_stripe = led_stripe_ok
            if self.has_led_stripe:
                if led_stripe is None:
                    led_stripe = LedStripe()
                self._led_stripe = led_stripe
            else:
                logger.warning('LED stripe activation failed')
                self._led_stripe = None
            self.has_buzzer = True
            self.buzzer = Buzzer(self.BUZ)
        elif self.hardware == "AlphaBot":
            self.has_speed = True
            self.has_ir_line = False
            self.has_ir_front = False
            self.has_led_stripe = False
            self._led_stripe = None  # type: LedStripe
            self.has_buzzer = False
            self.buzzer = None  # type: Buzzer

        if bluetooth_ok:
            self.bluetooth = _bluetooth
        else:
            self.bluetooth = None

        self._blink_thread = None  # type: ThreadUnit

        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)

        # SENSORS

        # Ultra-sonic distance estimation
        if self.hardware == "AlphaBot2":
            self.TRIG = 22
            self.ECHO = 27
        else:
            self.TRIG = 17
            self.ECHO = 5
        GPIO.setup(self.TRIG, GPIO.OUT, initial=GPIO.LOW)
        GPIO.setup(self.ECHO, GPIO.IN)

        # Check whether ultra-sound module is present
        self.has_ultra_sound = self.ultra_sound_distance() is not None

        # Store whether Bluetooth libraries were loaded successfully
        self.has_bt_rssi = bluetooth_ok

        # Store whether UWB libraries were loaded successfully
        self.has_uwb_module = uwb_ok

        # A/D conversion
        if self.has_ultra_sound and self.hardware == "AlphaBot":
            # on AlphaBot ECHO and CS both connect to Raspberry pin 5!!
            # when using ultra sound it is necessary to replace the AlphaBot
            # jumper between CS and pin 5 with a custom wire between CS and
            # pin 18 in order to have the tracer running (necessary for
            # analog front infra-red and line tracking)
            self.CS = 18
        else:
            self.CS = 5
        self.Clock = 25
        self.Address = 24
        self.DataOut = 23
        GPIO.setup(self.Clock, GPIO.OUT)
        GPIO.setup(self.Address, GPIO.OUT)
        GPIO.setup(self.CS, GPIO.OUT)
        GPIO.setup(self.DataOut, GPIO.IN, GPIO.PUD_UP)

        # Analog reading of infra-red sensors
        self.tracer_all_addresses = {IRP.Line1: 0,
                                     IRP.Line2: 1,
                                     IRP.Line3: 2,
                                     IRP.Line4: 3,
                                     IRP.Line5: 4,
                                     IRP.Left: 8,
                                     IRP.Right: 9,
                                     IRP.Battery: 10}
        self.tracer_front_addresses = [8, 9]
        self.tracer_line_addresses = [0, 1, 2, 3, 4]
        self.tracer_ports = IRP.Void

        # initialized in self._set_tracer_ports
        self._tracer_addresses = []
        self._front_indices = []
        self._line_indices = []
        self._battery_index = None

        self.led_stripe_thread = None
        self._led_off_color = None  # attribute to set the fall back
        # color for blinking led
        if self.hardware == "AlphaBot":
            # # Speed detection
            self.SDL, self.SDR = 7, 8
            self.buttons = [gpiozero.Button(self.SDL, pull_up=True),
                            gpiozero.Button(self.SDR, pull_up=True)]
            # if i use a loop instead of the 2 lines below, both function will
            # have argument 1, instead of 0 and 1!! is this a bug?
            self.buttons[0].when_pressed = lambda: self.motion_event(0)
            self.buttons[1].when_pressed = lambda: self.motion_event(1)
            self.speed_events = (deque(maxlen=_speed_buffer_size),
                                 deque(maxlen=_speed_buffer_size))
        else:
            self.s_button = gpiozero.Button(7)
            self.a_button = gpiozero.Button(8)
            self.b_button = gpiozero.Button(9)
            self.c_button = gpiozero.Button(10)
            self.d_button = gpiozero.Button(11)
            self.buttons = [self.s_button, self.a_button, self.b_button,
                            self.c_button, self.d_button]

        # Bluetooth RSSI (mac address to track)
        self.has_bt_rssi = bluetooth_ok
        if not bluetooth_ok:
            logger.warning('Bluetooth libraries not installed')
        # list of Bluetooth devices
        self.BT_devices = None
        self.BT_addr = None
        # activate UWB (this creates self.UWB_serial) but do not use yet
        self.UWB_serial = None
        # if uwb_ok:
        #     self._activate_uwb()
        self._use_UWB = False

        # Accelerometer
        if uwb_ok:
            self.Accelerometer = None  # TODO Change None, to UWB use acc
        self.has_accelerometer = accelerometer_ok
        self._use_accelerometer = False

        # Camera
        try:
            self.camera = PiCamera()
            self.has_camera = True
        except Exception as e:
            self.camera = None  # type: PiCamera
            self.has_camera = False
            logger.exception(e)
        # (camera will be used only upon depend)
        self._use_camera = False
        self._use_ultrasound = True
        self._camera_small_res = (64, 48)
        self._camera_downsample = None
        self._camera_subfield = False
        self._camera_roi_x = [0, self._camera_small_res[0]]
        self._camera_roi_y = [0, self._camera_small_res[1]]

        self._stop_after_delay_thread = None  # type: threading.Thread
        self._stop_after_delay_canceled = None

        # MOTOR
        if self.hardware == 'AlphaBot':
            self.IN1, self.IN2 = 12, 13
        elif self.hardware == 'AlphaBot2':
            self.IN1, self.IN2 = 13, 12
        self.ENA = 6
        self.IN3, self.IN4, self.ENB = 20, 21, 26
        GPIO.setup(self.IN1, GPIO.OUT)
        GPIO.setup(self.IN2, GPIO.OUT)
        GPIO.setup(self.IN3, GPIO.OUT)
        GPIO.setup(self.IN4, GPIO.OUT)
        GPIO.setup(self.ENA, GPIO.OUT)
        GPIO.setup(self.ENB, GPIO.OUT)
        self._current_motor = [0, 0]
        self.PWMA = GPIO.PWM(self.ENA, 500)
        self.PWMB = GPIO.PWM(self.ENB, 500)
        self.PWMA.start(50)
        self.PWMB.start(50)

        # Reset motor and buzzer in case they a previous program run did not
        # reset them correctly, but do not turn off leds as they are
        # controlled by the BootingLeds object
        self.set_motor(0, 0)
        self.set_buzzer(0)

        # left/right bias
        bias = self.get_robot_info(LEFT_RIGHT_BIAS)
        if bias is None:
            bias = 1.
        else:
            logger.info('load left/right bias: {}'.format(bias))
        self._left_right_bias = bias

    def set_leds_off_color(self, value):  # we set the default leds color
        self._led_off_color = value

    # robot parameters saved in json file
    @staticmethod
    def get_robot_info(name, default=None):
        return RobotInfo.get_robot_info(name, default)

    @staticmethod
    def update_robot_info(name, value):
        RobotInfo.update_robot_info(name, value)
        logger.info("Updated {} with value {}".format(name, value))

    @property
    def color_name(self):
        return self.get_robot_info(COLOR)

    # returns the shell color as a (4,3) matrix
    @property
    def color(self):
        return colors_by_name(self.color_name)

    @color.setter
    def color(self, value):
        self.update_robot_info(COLOR, value)

    @property
    def left_right_bias(self):
        # simple accessor
        return self._left_right_bias

    def update_left_right_bias(self, value=1):
        # we can not update the attribute without updating the file
        # so we group them
        self._left_right_bias = value
        self.update_robot_info(LEFT_RIGHT_BIAS, self._left_right_bias)

    def _read_hardware(self):
        command = ["cat", "/proc/device-tree/model"]
        try:
            output = subprocess.check_output(
                command, stderr=subprocess.STDOUT
            ).decode()
            if 'Pi Zero W' in output\
                    or "Pi Zero 2 W" in output\
                    or "Pi 3" in output or "Pi 4" in output:
                self.hardware = "AlphaBot2"
            else:
                self.hardware = "AlphaBot"
        except Exception as e:
            logger.warning(
                'Command "%s" failed, assumed robot is AlphaBot' % command)
            logger.exception(e)
            self.hardware = "AlphaBot"

        logger.info("robot is : {}".format(self.hardware))

    def cleanup(self):
        logger.info("Cleanup alpha bot")
        # stop robot
        self.stop()
        # stop accelerometer
        if self._use_accelerometer:
            self.use_accelerometer(False)
        # stop camera
        if self._use_camera:
            self.use_camera(False)
        if self.camera is not None:
            self.camera.close()
        # stop Bluetooth
        if self.BT_addr:
            self.use_BT_rssi(None)
        # close individual gpio devices
        for button in self.buttons:
            button.close()
        self.buttons = []
        self.buzzer.close()
        # general gpio cleanup
        pin_factory = gpiozero.devices.Device.pin_factory
        if pin_factory is not None:
            # closing the "pin_factory" occurs at program exit and fails if
            # GPIO.cleanup() was called before, even if individual devices
            # were closed as we did above, so we explicitly perform it before
            # calling GPIO.cleanup()
            pin_factory.close()
        GPIO.cleanup()

    # Get all sensors data
    def get_sensor_data(self):
        t0 = time.time()
        s = Communication.SensorData()

        if self._tracer_addresses:
            # collect measurements
            tracer_data = self._get_analog_data(self._tracer_addresses)
            # front infra-red
            s.IR_front = tracer_data[self._front_indices]
            # line-tracking infra-red
            s.IR_line = tracer_data[self._line_indices]
            # battery level
            if self._battery_index is not None:
                s.battery_level = tracer_data[self._battery_index]

        # Speed
        if self.has_speed:
            s.speed = self.get_speed()

        # Accelerometer
        if self._use_accelerometer:
            s.accelerometer = self.Accelerometer.collect_measurements()

        # BT distance
        if (bluetooth_ok and self.BT_addr is not None
                and self.bluetooth_thread is not None):
            bluetooth_data = self.bluetooth_thread.collect_measurements()

            if bluetooth_data.size > 0:
                timestamps = bluetooth_data[:, 0]
                values = bluetooth_data[:, 1:]
            else:
                # array is empty: make empty list of timestamps and values
                timestamps = np.zeros(0)
                values = np.zeros((0, 10))

            s.BT_rssi = Communication.TimeStampedData(timestamps, values[:])

        # UWB distance
        if self._use_UWB:
            s.UWB_value = self.UWB_serial.read_uwb()

        # Ultra-sound
        if self.has_ultra_sound and self._use_ultrasound:
            s.US_distance = self.ultra_sound_distance(ignore_error=True)

        # Camera
        elapsed_camera = None
        if self._use_camera:
            t1 = time.time()
            s.camera = self.get_picture()
            elapsed_camera = time.time() - t1

        # CPU temperature
        s.temperature = gpiozero.CPUTemperature().temperature

        # Current robot time
        s.time = time.time()

        # Display elapsed times
        elapsed_str = 'get_sensor_data elapsed: {:.3f}'.format(s.time - t0)
        if self._use_camera:
            elapsed_str += ' (camera: {:.3f})'.format(elapsed_camera)
        logger.debug(elapsed_str)

        # print('got sensor data in %.1fms' % ((time.time() - t) * 1e3))
        return s

    # Individual sensors
    def get_digital_IR(self):
        return GPIO.input(self.DL), GPIO.input(self.DR)

    def _get_analog_data(self, addresses: list, nrepeat=1) -> np.array:
        # addresses is empty, but this case should be handled separately
        DEBUG = False
        if DEBUG:
            bits = [0] * 12

        # we don't know why this happens sometimes...
        if len(addresses) == 0:
            logger.warning(
                'no address in _get_analog_data! {} {}'.format(
                    addresses, self._tracer_addresses
                )
            )
            return np.zeros(0)

        # Prepare container for values
        naddress = len(addresses)
        values = np.zeros(naddress * nrepeat, 'uint16')

        # Prepare list of addresses to ask conversion for
        addresses = np.array(addresses).repeat(nrepeat)

        # Loop: one more loop repeat because the address set at loop k is
        # sampled at loop (k+1)
        for k_addr in range(naddress * nrepeat + 1):
            GPIO.output(self.CS, GPIO.LOW)
            if k_addr < naddress * nrepeat:
                address = addresses[k_addr]
                if DEBUG:
                    logger.info('address {}'.format(address))

            # Sample 10 bits value and set address for next time
            for i in range(12):
                # Set address
                if i < 4:
                    # noinspection PyUnboundLocalVariable
                    if address is not None:
                        if (address >> (3 - i)) & 0x01:
                            GPIO.output(self.Address, GPIO.HIGH)
                        else:
                            GPIO.output(self.Address, GPIO.LOW)
                    else:
                        GPIO.output(self.Address, GPIO.LOW)
                else:
                    GPIO.output(self.Address, GPIO.LOW)

                # Collect the currently sampled bit
                if k_addr > 0 and i < 10:
                    values[k_addr - 1] <<= 1
                    if GPIO.input(self.DataOut):
                        values[k_addr - 1] |= 0x01
                if DEBUG:
                    bits[i] = GPIO.input(self.DataOut)

                # Trigger reading the address bit from ADDRESS
                GPIO.output(self.Clock, GPIO.HIGH)
                # time.sleep(250e-9)  # min pulse dur 190ns, max freq 2.1MHz

                # Trigger shift to the next bit on DATAOUT
                GPIO.output(self.Clock, GPIO.LOW)
                # time.sleep(250e-9)  # min pulse dur 190ns, max freq 2.1MHz

            # Wait to ensure conversion finished (takes 9.5us)
            time.sleep(1e-5)
            if DEBUG:
                logger.debug(''.join([str(x) for x in bits]))

            # Disable TCL1543 board
            GPIO.output(self.CS, GPIO.HIGH)

        # Output
        if nrepeat > 1:
            # Average over repeats
            values = values.reshape((naddress, nrepeat))
            values = values.mean(axis=1, dtype='float')
        return values

    def motion_event(self, side):
        self.speed_events[side].append(time.time())

    def get_speed(self):
        speed = [0] * 2
        for i in range(2):
            q = np.array(self.speed_events[i])
            q = q[time.time() - q < _speed_integrate_time]
            if len(q) < 2:
                # less than 2 recent events, robot is not moving
                speed[i] = 0
            else:
                speed[i] = (len(q) - 1) / (q[-1] - q[0])  # unit: events/second
        return speed

    def is_moving(self):
        return self.get_speed() > 0

    def ultra_sound_distance(self, timeout=0.8, ignore_error=False):
        GPIO.output(self.TRIG, GPIO.HIGH)
        time.sleep(0.00001)  # ne serait-il pas impossible que l'echo
        # n'arrive pendant ce sleep ?
        GPIO.output(self.TRIG, GPIO.LOW)
        tick = time.time()
        while not GPIO.input(self.ECHO):
            if time.time() > tick + timeout:
                logger.warning('No ultra-sound module detected')
                return timeout if ignore_error else None
        t1 = time.time()
        tick = time.time()
        while GPIO.input(self.ECHO):
            if time.time() > tick + timeout:
                logger.info('No echo detected')
                return timeout if ignore_error else None
        t2 = time.time()
        distance = (t2 - t1) * 34300 / 2
        return distance

    def get_bt_devices(self):
        logger.info("scanning bluetooth devices...")
        self.BT_devices = Bluetooth_Scan.scan_devices()
        logger.info('done')
        return self.BT_devices

    def _activate_uwb(self, activate=True):
        if activate:
            self.UWB_serial = UWB.CommunicationSerial()
            activated = self.UWB_serial.enable_uwb(True)
            if not activated:
                logger.warning("UWB activation failed")
        else:
            if self.UWB_serial:
                deactivated = self.UWB_serial.enable_uwb(False)
                if not deactivated:
                    logger.warning("UWB deactivation failed")
                self.UWB_serial = None

    def use_UWB(self, value):
        if self.UWB_serial:
            self.UWB_serial.use_uwb(value)
            self._use_UWB = value
        elif value:
            logger.warning('Cannot use UWB, UWB was not activated.')

    def _set_tracer_ports(self, ports: IRP):
        # Set which ports the tracer will read from. This includes front
        # infra-red ports and line-tracking infra-red ports. The list of the
        # used ports is encoded in the IRP flag 'ports' and will be
        # converted to a list of addresses.

        # Convert 0 or False or None to IRP.Void
        if not ports:
            ports = IRP.Void
        ports = ports | IRP.Battery
        self.tracer_ports = ports

        # convert ports flag into list of addresses
        # (below, p & ports means "p is inside the set of ports described by
        # flag ports")
        self._tracer_addresses = [addr
                                  for p, addr in
                                  self.tracer_all_addresses.items()
                                  if p & ports]
        self._tracer_addresses.sort()

        if self._tracer_addresses:
            self._front_indices = [
                i for i, addr in enumerate(self._tracer_addresses)
                if addr in self.tracer_front_addresses
            ]
            self._line_indices = [
                i for i, addr in enumerate(self._tracer_addresses)
                if addr in self.tracer_line_addresses
            ]
            try:
                self._battery_index = self._tracer_addresses.index(
                    self.tracer_all_addresses[IRP.Battery]
                )
            except ValueError:
                self._battery_index = None

    def use_IR_front(self, value: bool):
        # Set whether reading values for front infra-red sensors:
        # we build a ports flag that overrides the current setting for these
        # sensors, and preserves those of other sensors
        ports = self.tracer_ports
        if value:
            ports = ports | IRP.Front
        else:
            ports = ports & ~IRP.Front
        self._set_tracer_ports(ports)

    def use_IR_line(self, value: IRP):
        # Set whether reading values for line-tracking infra-red sensors:
        # we build a ports flag that overrides the current setting for these
        # sensors, and preserves those of other sensors
        ports_line = value & IRP.Line
        ports_other = self.tracer_ports & ~IRP.Line
        ports = ports_other | ports_line
        self._set_tracer_ports(ports)

    def use_BT_rssi(self, mac_addr):
        if not bluetooth_ok:
            # no bluetooth
            if mac_addr:
                raise RuntimeError('Bluetooth is not available.')
            return

        # Start or stop using thread
        if mac_addr:
            if self.bluetooth_thread is not None:
                self.bluetooth_thread = None

            self.bluetooth.enable_socket(False)
            self.BT_addr = mac_addr
            self.bluetooth_thread = ThreadUnit(
                lambda: self.bluetooth.pop_bt_rssi(1, mac_addr),
                read_period=1e-2, debug=False)
            self.bluetooth.enable_socket(True)
            self.bluetooth_thread.start()
        else:
            if self.bluetooth_thread is not None:
                self.bluetooth.enable_socket(False)
                self.bluetooth_thread.stop()
                self.bluetooth_thread = None

    def use_accelerometer(self, read_period):
        self._use_accelerometer = bool(read_period)
        if self._use_accelerometer:
            self.Accelerometer.start(read_period)
        else:
            self.Accelerometer.stop()

    def set_leds(self, value):
        if not self.has_led_stripe or self._led_stripe is None:
            return
        # stop blinking if any
        if self._blink_thread is not None:
            logger.debug("stop led blinking")
            self._blink_thread.stop(blocking=False)
            self._blink_thread = None
        # set color value(s) to LEDs
        value = np.broadcast_to(value, (4, 3))
        if self._led_stripe is not None:
            self._led_stripe.set_led_colors(value)

    def _blink_once(self, color, duration):
        # keep this method private because, contrary to set_leds, it doesn't
        # stop the blinking thread
        color = np.broadcast_to(color, (4, 3))
        if self._led_stripe is None:
            return
        self._led_stripe.set_led_colors(color)
        time.sleep(duration)
        if self._led_off_color is None:
            fallback_color = np.zeros((4, 3))
        else:  # if the fallback color has changed, the leds are set to the new
            # color instead of turn off
            fallback_color = np.broadcast_to(self._led_off_color, (4, 3))
        self._led_stripe.set_led_colors(fallback_color)

    def blink_leds(self, color=.5, period=1, on=.1, n_repeat=None):
        if not self.has_led_stripe:
            logger.debug("This robot does not have a led stripe.")
            return
        if period == 0:
            if n_repeat and n_repeat != 1:
                raise ValueError('If blink_leds period is zero, number of '
                                 'repeats can only be 1')
            n_repeat = 1
        if self._blink_thread is not None:
            self._blink_thread.stop()
        self._blink_thread = ThreadUnit(
            lambda: self._blink_once(color, on), read_period=period,
            n_repeat=n_repeat)
        self._blink_thread.start()

    @property
    def camera_subfield_values(self):
        return self._camera_subfield, self._camera_roi_x, self._camera_roi_y

    @camera_subfield_values.setter
    def camera_subfield_values(self, values):
        logger.info("subfield values set to {}".format(values))
        if values is not None:
            assert not self._camera_compress
            self._camera_subfield, self._camera_roi_x, self._camera_roi_y = values
        else:
            self._camera_subfield = False

    def use_ultrasound(self, value):
        self._use_ultrasound = value

    def use_camera(self, resolution):
        self.camera_subfield_values = None

        prev_value = self._use_camera
        self._use_camera = bool(resolution)

        if not self._use_camera:
            if prev_value:
                self.camera.stop_preview()
            return
        elif not self.has_camera:
            raise Exception('AlphaBot has no camera to use')

        if prev_value:
            self.camera.stop_preview()

        # values might be too low for the camera
        resolution = np.array(resolution)
        self._camera_resolution = resolution
        if all(resolution <= self._camera_small_res):
            # no jpeg compression, take pictures at resolution _small_res
            # and further bin them if necessary
            self._camera_downsample = int(
                max(np.ceil(np.array(self._camera_small_res) / resolution))
            )
            self._camera_compress = False
        else:
            self._camera_downsample = 1
            self._camera_compress = True

        res_raw = tuple(resolution * self._camera_downsample)
        self.camera.resolution = res_raw

        # # fix camera settings, see #
        # https://picamera.readthedocs.io/en/release-1.13/recipes1.html
        # #capturing-consistent-images
        #
        # # Set ISO to the desired value
        # self.camera.iso = 100
        # # Wait for the automatic gain control to settle
        # time.sleep(2)
        # # Now fix the values
        # self.camera.shutter_speed = self.camera.exposure_speed
        # self.camera.exposure_mode = 'off'
        # g = self.camera.awb_gains
        # self.camera.awb_mode = 'off'
        # self.camera.awb_gains = g

        if self._camera_compress:
            self._camera_raw = io.BytesIO()
        else:
            self._camera_raw = np.zeros((res_raw[1], res_raw[0], 3),
                                        dtype=np.uint8)

    def get_picture(self):
        if self._camera_compress:
            self._camera_raw.seek(0)
            self.camera.capture(self._camera_raw, 'jpeg', use_video_port=True)
            view = self._camera_raw.getbuffer()[:self._camera_raw.tell()]
            return view.tobytes()
        else:
            self.camera.capture(self._camera_raw, 'rgb', use_video_port=True)

            # downsampling
            image = self._camera_raw
            if self._camera_downsample:
                res_x, res_y = self._camera_resolution
                max_x, max_y = self._camera_small_res
                bin_x = bin_y = self._camera_downsample
                subfield, roi_x, roi_y = self.camera_subfield_values
                # cannot have subfield without downsample
                if subfield:
                    # slice the correct portion of the image
                    slice_x = [x * max_x // 100 for x in roi_x]
                    slice_y = [y * max_y // 100 for y in roi_y]
                    image = image[slice(*slice_y, 1), slice(*slice_x, 1)]
                    # get the bins
                    # (sub_size[%] / current_size[px]) * (max_size[px] / 100[%])
                    # analytically correct
                    # e.g. (50 / 16) * (64 /100) = (50*64)//(16*100) =
                    # 50.2^6//50.2^5 = 2^(6-5) = 2 for half the width
                    bin_x = (roi_x[1] - roi_x[0]) * max_x // (res_x * 100)
                    bin_y = (roi_y[1] - roi_y[0]) * max_y // (res_y * 100)
                image = np.reshape(image, [res_y, bin_y, res_x, bin_x, -1])
                image = np.mean(image, axis=(1, 3)).astype(np.uint8)
            return image

    # Motor command
    def stop(self):
        # stop motors
        self.set_motor(0, 0)
        # stop buzzer
        self.set_buzzer(0)

    # Function that will be executed in a separated thread to stop the robot
    # after a given duration
    def thread_delay(self, duration):
        time.sleep(duration)
        if not self._stop_after_delay_canceled:
            logger.info('stopping robot before a new command was received')
            self.stop()
        else:
            logger.info('received next command, not stopping robot')

    def set_motor_and_buzzer(self, left, right, buzz, duration):
        if self._stop_after_delay_thread is not None:
            # Cancel previous instruction to stop the robot after
            # a given duration
            self._stop_after_delay_canceled = True
            # Wait for the motor command to finish
            self._stop_after_delay_thread.join()

        # Set motor
        self.set_motor(left, right)

        # Set buzzer
        self.set_buzzer(buzz)

        # Start a thread that, if needed, will automatically stop the robot
        # after the requested duration
        if duration is not None:
            self._stop_after_delay_thread = threading.Thread(
                target=self.thread_delay, args=(duration,))
            self._stop_after_delay_canceled = False
            self._stop_after_delay_thread.start()

    def set_motor(self, left, right):
        # When a motor inverts direction, mark a brief stop to let it breathe
        breathe = False
        if left * self._current_motor[0] < 0:
            GPIO.output(self.IN1, GPIO.LOW)
            GPIO.output(self.IN2, GPIO.LOW)
            breathe = True
        if right * self._current_motor[1] < 0:
            GPIO.output(self.IN3, GPIO.LOW)
            GPIO.output(self.IN4, GPIO.LOW)
            breathe = True
        if breathe:
            logger.info('breathe')
            time.sleep(.03)

        # Set left motor (beware, forward left wheel motion makes the robot
        # turn right)
        if left == self._current_motor[0]:
            # no change for left motor
            pass
        elif left == 0:
            GPIO.output(self.IN1, GPIO.LOW)
            GPIO.output(self.IN2, GPIO.LOW)
        elif left > 0:
            GPIO.output(self.IN2, GPIO.LOW)
            GPIO.output(self.IN1, GPIO.HIGH)
            self.PWMA.ChangeDutyCycle(min((left, 100)))
        else:
            GPIO.output(self.IN1, GPIO.LOW)
            GPIO.output(self.IN2, GPIO.HIGH)
            self.PWMA.ChangeDutyCycle(min((-left, 100)))

        # Set right motor (and vice-versa...)
        if right == self._current_motor[1]:
            # no change for right motor
            pass
        elif right == 0:
            GPIO.output(self.IN3, GPIO.LOW)
            GPIO.output(self.IN4, GPIO.LOW)
        elif right > 0:
            GPIO.output(self.IN3, GPIO.LOW)
            GPIO.output(self.IN4, GPIO.HIGH)
            self.PWMB.ChangeDutyCycle(min((right, 100)))
        else:
            GPIO.output(self.IN4, GPIO.LOW)
            GPIO.output(self.IN3, GPIO.HIGH)
            self.PWMB.ChangeDutyCycle(min((-right, 100)))

        # Memorize current motor
        self._current_motor = [left, right]

    def set_buzzer(self, tone=0):
        if self.has_buzzer:
            self.buzzer.buz(tone)

    def play_jingle(self, reverse=False):
        notes = [440, 660, 880, 660]
        if reverse:
            notes = notes[::-1]

        def play():
            for note in notes:
                self.set_buzzer(note)
                time.sleep(0.2)
            self.set_buzzer(0)

        jingle_thread = threading.Thread(target=play)
        jingle_thread.start()

    def jiggle(self, n=10, blocking=False):
        if blocking:
            self._jiggle(n)
        else:
            # execute dance in thread to not block execution
            threading.Thread(
                target=self._jiggle, args=(n,),
                name="wifi").start()

    def _jiggle(self, n):
        # execute small dance in thread to not block execution
        for i in range(n):
            GPIO.output(self.IN1, GPIO.HIGH)
            GPIO.output(self.IN3, GPIO.HIGH)
            time.sleep(0.02)
            GPIO.output(self.IN1, GPIO.LOW)
            GPIO.output(self.IN3, GPIO.LOW)
            time.sleep(0.02)
            GPIO.output(self.IN2, GPIO.HIGH)
            GPIO.output(self.IN4, GPIO.HIGH)
            time.sleep(0.02)
            GPIO.output(self.IN2, GPIO.LOW)
            GPIO.output(self.IN4, GPIO.LOW)
            time.sleep(0.02)

    def __del__(self):
        self.cleanup()


if __name__ == '__main__':
    # I suppose that the code is executed from the robot, and that we are
    # testing the robot
    pass
