import time
from threading import Thread
from LedStripe import colors_by_name
from RobotLogger import robot_logger as logger


class BootingLeds:
    off_color = [0., 0., 0.]

    def __init__(self, led_stripe, booting_colors=None):
        self._led_stripe = led_stripe
        booting_colors = booting_colors or [0., 1., 0.]
        if isinstance(booting_colors, str):
            booting_colors = colors_by_name(booting_colors)
        if len(booting_colors) == 3:
            # defined a single color -> copy it to make 4 colors
            booting_colors = [booting_colors] * 4
        self.booting_colors = booting_colors
        self.color_thread = None  # type: Thread
        self.is_alive = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return True

    def _set_led_colors(self, colors):
        if self._led_stripe is None:
            logger.debug("self._led_stripe is None")
            return
        if isinstance(colors, str):
            colors = colors_by_name(colors)
        self._led_stripe.set_led_colors(colors)

    def booting_lights(self):
        led_state = 0
        while self.is_alive and self._led_stripe is not None:
            led_state = (led_state + 1) % 6  # ping-pong light
            if led_state < 4:
                lighted_led = led_state
            else:
                lighted_led = 6 - led_state
            color = [
                self.booting_colors[i] if i == lighted_led
                else self.off_color
                for i in range(4)]
            self._set_led_colors(color)
            time.sleep(0.3)

    def start(self):
        self.color_thread = Thread(target=self.booting_lights)
        self.is_alive = True
        self.color_thread.start()

    def set_booting_colors(self, r=0., g=1., b=0.):
        self.booting_colors = [r, g, b]

    def done(self):  # Alias for stop
        self.stop()

    def stop(self):
        self.is_alive = False
        # wait for thread to terminate
        if self.color_thread.is_alive():
            self.color_thread.join()
            self.color_thread = None
        # turn leds off
        self._set_led_colors([self.off_color] * 4)

    def __del__(self):
        self._set_led_colors([self.off_color] * 4)
