import time
import os
from contextlib import contextmanager
from enum import IntEnum
import sys
import numpy as np

from AlphaBot import AlphaBot
from RobotLogger import robot_logger as logger
from RobotKNN import KNN
from LedStripe import LedStripe
from BootingLeds import BootingLeds as WaitingLed

led_stripe = LedStripe()
waiting_led = WaitingLed(led_stripe, booting_color=[1, 0, 1])
waiting_led.start()

PAUSE_BEFORE_IMAGE = 0
SENSOR_AHEAD = .15
LOOP_TIME = .2
MAX_SPEED = 74
ACCELERATION = 2
TURN_FACTOR = .5
BACK_WHILE_TURN = .15


class Mode(IntEnum):
    reset = 0b0000
    controlled = 0b0001
    autonomous = 0b0010
    speed = 0b0100
    algo = 0b1000
    controlled_autonomous = controlled | autonomous  # 0b0011
    speed_autonomous = speed | autonomous            # 0b0101
    algo_autonomous = algo | autonomous              # 0b1001

    @classmethod
    def next(cls, mode):
        if mode == cls.controlled:
            return cls.controlled_autonomous
        elif mode == cls.controlled_autonomous:
            return cls.speed_autonomous
        elif mode == cls.speed_autonomous:
            return cls.algo_autonomous
        elif mode == cls.algo_autonomous:
            return cls.controlled
        else:
            raise ValueError


class Client:
    _speed = 40
    _is_running = True
    _mode = Mode.controlled

    def __init__(self):
        if len(sys.argv) > 1 and sys.argv[1].isdigit():
            logger.info("Killing parent process")
            os.kill(int(sys.argv[1]), 9)
            time.sleep(3)  # wait for camera to be released
        self.robot = AlphaBot()
        self.robot.use_ultrasound(False)
        try:
            self.robot.use_camera((16, 12))
        except Exception as e:
            logger.exception(e)
            waiting_led.done()
            self.robot.set_leds([[0., 1., 0.]] * 4)
            sys.exit(1)
        self.robot.camera_subfield_values = (True, (0, 100), (25, 100))
        self._camera_locked = False
        self._last_image = None
        self._setup_algorithm()
        self.has_reset = False

        def reset():
            self.has_reset = True
            self.robot.jiggle(5)
            self._setup_algorithm()

        self.robot.s_button.hold_time = 2
        self.robot.s_button.when_held = reset

        self.algorithm.load_buffers_from_file()

        self.robot.s_button.when_released = self.switch_mode
        waiting_led.done()
        self._set_mode()
        self.robot.jiggle(10)
        while self._is_running:
            if self._mode & Mode.autonomous:
                # stop before taking picture..., or take picture without
                # stopping, some time in advance before next decision
                if not SENSOR_AHEAD:
                    self.stop()
                    if PAUSE_BEFORE_IMAGE:
                        time.sleep(PAUSE_BEFORE_IMAGE)
                else:
                    assert not PAUSE_BEFORE_IMAGE
                    tic = time.perf_counter()

                # take new decision
                data = self._get_image()
                decision = self.algorithm.get_and_memorize_decision(
                    data.ravel())
                if decision is None:
                    left, right = 0, 0
                else:
                    left, right = decision

                # wait for previous action to complete
                if SENSOR_AHEAD:
                    tac = time.perf_counter()
                    if tac - tic < SENSOR_AHEAD:
                        time.sleep(SENSOR_AHEAD - tac + tic)
                    else:
                        print('taking decision took more time than expected, increase SENSOR_AHEAD value')

                # execute new action
                tic = time.perf_counter()
                self._move(left, right)
                # self.algorithm.save_buffers_to_file()
                tac = time.perf_counter()
                if tac - tic < LOOP_TIME - SENSOR_AHEAD:
                    time.sleep(LOOP_TIME - SENSOR_AHEAD - tac + tic)
                else:
                    print('setting motors took more time than expected, '
                          'increase LOOP_TIME')
            else:
                self.stop()
                time.sleep(LOOP_TIME)

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value):
        self._mode = value
        self._set_mode()

    def switch_mode(self):
        if not self.has_reset:
            self.mode = Mode.next(self.mode)
        else:
            self.has_reset = False

    def _setup_algorithm(self):
        self.stop()
        current_image = self._get_image()
        self.algorithm = KNN(
            self.robot, current_image.size, 2, 1, True, None, LOOP_TIME
        )

    def _get_image(self):
        if not self._camera_locked:
            with self.lock():
                camera = self.robot.get_sensor_data().camera
                camera = camera.astype(np.float32)
                # Concert RGB to Luminance, Yellow vs. Blue, Red vs. Green
                r, g, b = np.moveaxis(camera, -1, 0)
                W = (r + g + b) / 3  # btw 0 (black) and 1 (white)
                Y = (r + g - 2 * b) / 2  # btw -.5 (blue) and 1 (yellow)
                R = r - g  # btw -1 (green) and 1 (red)
                # In fact we are not interested in Yellow vs. Blue, and we
                # must emphasize more the Red vs. Green
                camera = np.stack((W, 2*R), axis=2)

            self._last_image = camera
            return self._last_image
        else:
            print('camera locked, returning previous image!')
            return self._last_image

    @contextmanager
    def lock(self):
        self._camera_locked = True
        yield
        self._camera_locked = False

    @property
    def speed(self):
        return self._speed

    @speed.setter
    def speed(self, value):
        self._speed = value
        logger.info("speed is set to {}".format(self._speed))

    def increase_speed(self):
        if self.speed < MAX_SPEED:
            self.speed += ACCELERATION

    def decrease_speed(self):
        if self.speed > ACCELERATION:
            self.speed -= ACCELERATION

    def increase_bias(self):
        self.robot.update_left_right_bias(self.robot.left_right_bias + .01)

    def decrease_bias(self):
        self.robot.update_left_right_bias(self.robot.left_right_bias - .01)

    def forward(self):
        self._move_and_learn(1, 1)

    def backward(self):
        self._move_and_learn(-1, -1)

    def left(self):
        self._move_and_learn(
            -TURN_FACTOR - BACK_WHILE_TURN,
            TURN_FACTOR
        )

    def right(self):
        self._move_and_learn(
            TURN_FACTOR,
            -TURN_FACTOR - BACK_WHILE_TURN
        )

    def stop(self):
        self.robot.set_motor(0, 0)  # do not use robot.

    def _move_and_learn(self, left, right):
        data = self._get_image()
        tic = time.perf_counter()
        self._move(left, right)
        self.algorithm.learn((left, right), data.ravel())
        self.algorithm.save_buffers_to_file()
        tac = time.perf_counter()
        if tac - tic < LOOP_TIME:
            time.sleep(LOOP_TIME - tac + tic)

    def _move(self, left, right):
        self.robot.set_motor(
            left * self.speed * self.robot.left_right_bias,
            right * self.speed / self.robot.left_right_bias
        )

    def _set_mode(self):
        print('switched mode to', self._mode)
        if self._mode & Mode.controlled:
            b = int(bool(self._mode & Mode.autonomous))
            self.robot.set_leds([[b, 1., 1 - b]] * 4)

            self.robot.a_button.when_pressed = None
            self.robot.b_button.when_pressed = None
            self.robot.c_button.when_pressed = None
            self.robot.d_button.when_pressed = None

            self.robot.a_button.when_released = self.forward
            self.robot.b_button.when_released = self.right
            self.robot.c_button.when_released = self.left
            self.robot.d_button.when_released = self.backward
        elif self._mode & Mode.speed:
            self.robot.set_leds([[1., 0., 1.]] * 4)

            self.robot.a_button.when_pressed = None
            self.robot.b_button.when_pressed = None
            self.robot.c_button.when_pressed = None
            self.robot.d_button.when_pressed = None

            self.robot.a_button.when_released = self.increase_speed
            self.robot.b_button.when_released = self.increase_bias
            self.robot.c_button.when_released = self.decrease_bias
            self.robot.d_button.when_released = self.decrease_speed
        elif self._mode & Mode.algo:
            self.robot.set_leds([[1., 0., 0.]] * 4)

            self.robot.a_button.when_pressed = None
            self.robot.b_button.when_pressed = None
            self.robot.c_button.when_pressed = None
            self.robot.d_button.when_pressed = None

            self.robot.a_button.when_released = self.algorithm.increase_k
            self.robot.b_button.when_released = self.algorithm.save_buffers_to_file
            self.robot.c_button.when_released = None
            self.robot.d_button.when_released = self.algorithm.decrease_k
        elif self._mode & Mode.reset:
            self.robot.set_leds([[0., 1., 0.]] * 4)

            self.robot.a_button.when_pressed = None
            self.robot.b_button.when_pressed = None
            self.robot.c_button.when_pressed = None
            self.robot.d_button.when_pressed = None

            self.robot.a_button.when_released = None
            self.robot.b_button.when_released = None
            self.robot.c_button.when_released = None
            self.robot.d_button.when_released = None


if __name__ == '__main__':
    client = Client()
