# Do not import Utils because this code is executed both on the PC side and
# on the robot side (which does not have Utils.py)
import time
import pickle
import socket
import struct
import threading
# TODO: import a subset of Utils that can be imported on the robot as well
import traceback
from enum import IntEnum
from typing import Tuple, Sized
import os

import numpy as np

# the robot and PC doesn't import all the same file :


# For the PC side :
try:
    from Program.Utils import *
    from Program.Constants import *
    from Program.Parameters import connection_param
    from Environment.AlphAI import Upload_Robot_Code
    from Program.Errors import RobotConnectionLost


    class RobotCodeUploadedError(InternalError):
        def handle_error(self):
            # Switch back to fake robot
            connection_param.target_connected = False

            # Tell user to reconnect once robot has restarted
            robot_auto_restart = self.args[0]
            if robot_auto_restart:
                txt = tr('Up-to-date code has been uploaded and robot was '
                         'restarted.') + ' '
                txt += tr("If your robot does not restart by itself, please"
                          " restart it manually") + ". "
                txt += tr('Once it has finished restarting, reconnect to its '
                          'Wi-Fi and press "connected" again.')
                message_box(txt)
            else:
                message_box(
                    tr('Up-to-date code has been uploaded.') + ' '
                    + tr('Please turn robot off and on again.') + ' '
                    + tr('Once it has finished restarting, reconnect to its '
                         'Wi-Fi and press "connected" again.')
                )
            return True


    class ConnectionFailedError(InternalError):
        def handle_error(self):
            if connection_param.wifi_connection:
                info_msg = tr(
                    "Make sure the robot is on and flashing white, "
                    "and that you are connected to its Wi-Fi network"
                )
            elif len(connection_param.bt_mac_address) > 0:
                info_msg = tr(
                    "Make sure the robot is on and flashing white, "
                    "and you have selected the correct robot in the "
                    "bluetooth list"
                )
            else:
                info_msg = tr("Select Wi-Fi or Bluetooth in Connection menu")
            error_message(tr("Connection failed"), info_msg)
            return True


except ImportError:
    # For the robot side :
    # Occurs when code is executed on robot, but in this case
    # Upload_Robot_code and other stuff is not needed
    Upload_Robot_Code = connection_param = None
    import Utils_bluetooth_for_robot
    # get a logger for the robot as well
    from RobotLogger import robot_logger as logger


    class InvalidCaseError(Exception):
        pass

    class RobotConnectionLost(ConnectionError):
        pass


class ConnectionCheckOK(Exception):
    pass


class ConnectionCheckFail(Exception):
    pass


# Signal needs for translation
xtr = lambda s: s  # signals chains that will need translation
xtr2 = lambda s: s  # signals chains that will need double translation


def display_error(err):
    e_type = type(err).__name__
    msg = str(err)
    traceback.print_tb(err.__traceback__)
    logger.exception('Caught %s: %s' % (e_type, msg))


_wifi_port = 50034
_bt_port = 4  # bluetooth does not support 4 digit port.
# The first 3 are already in use on some raspberries

robot_code_version = "2.7.19"


class TracerPort(IntEnum):
    Void = 0

    Left = 1
    Right = 2

    Line1 = 4
    Line2 = 8
    Line3 = 16
    Line4 = 32
    Line5 = 64

    Front = Left | Right

    LineCenter = Line3
    LineMedium = Line2 | Line4
    LineExtrem = Line1 | Line5
    Line = LineCenter | LineMedium | LineExtrem

    All = Front | Line

    Battery = 128

    def n_sample(self):
        # number of bits equal to 1 in binary representation
        return sum([bool(self & 2 ** i) for i in range(7)])

    def __str__(self):
        if self == TracerPort.LineCenter:
            return xtr2('tracer_line_center')
        elif self == TracerPort.LineMedium:
            return xtr2('tracer_line_medium')
        elif self == TracerPort.LineExtrem:
            return xtr2('tracer_line_extrem')
        elif self == TracerPort.Line:
            return xtr2('tracer_line')
        else:
            return super(TracerPort, self).__str__()


class TimeStampedData:
    __slots__ = ['timestamps', 'values', 'n_per_sample']

    def __init__(self, times=None, val=None, n_per_sample=None):
        # two syntaxes are allowed: either define times and val, either
        # define n_per_sample
        if times is None and val is None:
            assert n_per_sample is not None
            self.n_per_sample = n_per_sample  # type: int
            self.timestamps = np.zeros(0)  # type: np.ndarray
            self.values = np.zeros((0, n_per_sample))  # type: np.ndarray
        else:
            assert times is not None \
                   and val is not None and n_per_sample is None
            if not isinstance(times, Sized):
                # only a single time point
                times, val = [times], [val]
            else:
                assert len(val) == len(times)
            self.timestamps = np.array(times)
            self.values = np.array(val)
            if self.values.ndim == 1:
                self.values = self.values[:, np.newaxis]
            self.n_per_sample = self.values.shape[1]

    @property
    def n_value(self):
        return len(self.timestamps)

    def add_values(self, val, subtract_start_time=0,
                   limit_time=None, limit_window=None):
        # type: (TimeStampedData, float, float, float) -> None
        timestamps = val.timestamps
        if subtract_start_time:
            # do not use a -= because this will modify the input values as well
            timestamps = timestamps - subtract_start_time
        self.timestamps = np.hstack((self.timestamps, timestamps))
        self.values = np.vstack((self.values, val.values))

        if limit_time or limit_window:
            self.throw_old(limit_time=limit_time, limit_window=limit_window)

    def throw_old(self, limit_time=None, limit_window=None):
        if limit_window and self.n_value:
            if limit_time:
                raise ValueError('limit_time and limit_window cannot be set '
                                 'at the same time')
            limit_time = self.timestamps[-1] - limit_window
        ok = self.timestamps > limit_time
        self.timestamps = self.timestamps[ok]
        self.values = self.values[ok, :]


class SensorData:
    IR_front = None  # type: TimeStampedData
    IR_thymio = None
    IR_line = None  # type: TimeStampedData
    speed = None  # type: Tuple[float, float]
    speed_ticks = None
    US_distance = None  # type: float
    camera = None  # type: np.ndarray
    BT_rssi = None  # type: TimeStampedData
    time = None  # type: float
    accelerometer = None  # TODO: put correct types
    UWB_value = None
    battery_level = None
    temperature = None
    fake_robot_state = None


# SOCKETS - The channel between robot and PC running AI consists of a single
# socket used to send and receive messages.
# PROTOCOL - It is always the PC that sends commands to the robot through the
# socket. The robot wait for commands. For some commands it responds with
# sensor data (sometimes after some specific waiting time) before listening for
# new commands. If the PC does not read after it issued a command that return
# sensor data, sensor data will accumulate and arrive with a delay. If on the
# contrary PC attempts to read in the socket without having sent a
# command to request data, a timeout will occur and cause disconnection.
# DISCONNECTION - If a disconnection occurs, for example because of a timeout,
# or because Wi-Fi disconnected, an exception will be risen that should be
# handled by the main program. In the case of the robot code, the main robot
# loop will initiate a new connection. In the case of the PC code, the state
# will switch from connected to disconnected.

# TIMEOUTS - For optimal user experience, the following timeouts are set
# before attempting reconnection

# Default timeout - when unspecified as argument for reading
TIMEOUT_DEFAULT = 15.0

# Default timeout for the robot = how long it waits before resetting the
# connection and waiting for a new hand_check.
TIMEOUT_ROBOT = 15

# When robot needs to wait longer for a user action such as a dialog window.
TIMEOUT_ROBOT_DIALOG = 60

# How long the PC wait for an answer when attempting to connect
TIMEOUT_PC_CONNECT = 5

# How long the PC waits for an answer when asking for e.g sensor data
TIMEOUT_PC_READ = 10.0

# Short timeout to allow PC to quickly check if new sensor data is available
TIMEOUT_PC_CHECK = 0.001

# - special: Bluetooth scan
TIMEOUT_BT_SCAN = 15


class BaseChannel:
    robot_is_fake = None

    def __init__(self):
        # Will be set upon init
        self.is_PC = False
        self.connected = False

        # Will be set upon hand check
        self.robot_hardware = None  # default for fake robot
        self.has_accelerometer = False
        self.has_ultra_sound = False
        self.has_camera = False
        self.has_bt_rssi = False
        self.has_uwb_module = False
        self.has_speed = False
        self.has_ir_front = False
        self.has_ir_line = False
        self.has_led_stripe = False

        self.joystick_thread = None  # type: ThreadUnit

    def _send_object(self, name, value):
        raise NotImplementedError

    def _read_object(self, wait=True, timeout=TIMEOUT_DEFAULT):
        raise NotImplementedError

    def send_sensor(self, name, value):
        if self.is_PC:
            raise Exception('PC cannot send sensor data')
        self._send_object(name, value)

    def send_command(self, name, value=None):
        if not self.is_PC:
            raise Exception('Robot cannot send motor data')
        return self._send_object(name, value)

    def read_sensor(self, wait=True) -> Tuple[str, SensorData]:
        if not self.is_PC:
            raise Exception('Robot cannot read sensor data')
        return self._read_object(wait, timeout=TIMEOUT_PC_READ)

    def read_command(self):
        if self.is_PC:
            raise Exception('PC cannot read motor data')
        return self._read_object(timeout=TIMEOUT_ROBOT)

    def scan_bluetooth_devices(self):
        logger.info("default (fake) Bluetooth scan")
        addr_list = ['AA:AA:AA:AA:AA:AA', 'BB:BB:BB:BB:BB:BB']
        return addr_list


class ConnectionStatus(IntEnum):
    NotConnected = 0
    Ready = 1
    Wifi = 2
    Bluetooth = 3

    def __str__(self):
        if self == ConnectionStatus.NotConnected:
            return "not connected"
        elif self == ConnectionStatus.Ready:
            return "waiting for connection"
        elif self == ConnectionStatus.Wifi:
            return "connected with Wi-Fi"
        elif self == ConnectionStatus.Bluetooth:
            return "connected with Bluetooth"
        else:
            return "invalid status"


class CommunicationChannel(BaseChannel):
    robot_is_fake = False

    def __init__(self, side, robot_ip='10.0.0.1', robot=None):
        super().__init__()

        # are we on the PC or robot side?
        self.side = side
        if side == 'PC':
            self.is_PC = True
            self.robot_ip = robot_ip
        elif side == 'robot':
            self.is_PC = False
            self.robot = robot
        else:
            raise Exception("side argument must be either 'PC' or 'robot'")
        self.connection_status = ConnectionStatus.NotConnected
        self.connection = None  # type: socket.socket
        self.wifi_socket = self.bt_socket = None  # type: socket.socket
        self.wifi_connection_thread = None  # type: threading.Thread
        self.bt_connection_thread = None # type: threading.Thread
        self.send_file = None
        self._lock = threading.RLock()  # reentrant lock to allow read and
        # writes while in check_hand function
        self.first_no_data_time = None

    def _wait_for_connection(self, protocol, connection_socket):
        try:
            connection, _ = connection_socket.accept()
        except OSError:
            # Happens when the socket is closed from the main thread
            connection = None
        if self.connection_status != ConnectionStatus.Ready:
            # Connection already established or main thread is out of the
            # waiting loop
            return
        if connection is not None:
            self.connection = connection
            self.connection.settimeout(TIMEOUT_ROBOT)
            self.connection_status = protocol  # Will cause main thread to
            # resume
        else:
            self.connection_status = ConnectionStatus.NotConnected  # will
            # cause check_hand to restart

    def check_hand_PC(self, check_only=False, b_mac=None):
        # we check the protocol to create the right socket+
        if check_only:
            protocol = ConnectionStatus.Bluetooth
        elif connection_param.wifi_connection:
            protocol = ConnectionStatus.Wifi
        else:
            protocol = ConnectionStatus.Bluetooth
        if self.connection is not None:
            self.connection.close()  # free used addresses
        if protocol == ConnectionStatus.Wifi:
            self.connection = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM
            )
            address = (self.robot_ip, _wifi_port)
        elif protocol == ConnectionStatus.Bluetooth:
            self.connection = socket.socket(
                socket.AF_BLUETOOTH,
                socket.SOCK_STREAM,
                socket.BTPROTO_RFCOMM
            )
            if b_mac is None:
                b_mac = connection_param.bt_mac_address
            address = (b_mac, _bt_port)
        else:
            raise InvalidCaseError
        self.connection.settimeout(TIMEOUT_PC_CONNECT)
        self.connection.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, 1
        )
        try:
            self.connection.connect(address)
            self.connected = True
            if not check_only:
                logger.info(' [connected]')
        except socket.timeout:
            self.connection.close()
            if check_only:
                raise ConnectionCheckFail
            else:
                logger.warning(' [connection failed]')
                raise ConnectionFailedError
        except Exception as err:
            self.connection.close()
            logger.warning("Unexpected exception encountered while "
                           "attempting to connect: "+str(err))
            raise err
        self.send_file = self.connection.makefile('wb')

        # read info
        try:
            name, info = self._read_object(timeout=TIMEOUT_PC_READ)
        except TypeError:  # In case of timeout, read_object will return None
            raise ConnectionError

        # check only
        if check_only:
            self._send_object('disconnect')
            return info

        # check robot code version
        try:
            remote_version = info['robot_code_version']
        except KeyError:
            remote_version = 1.0
        if remote_version != robot_code_version:
            self.robot_code_update(remote_version)  # Display a message box
            # and might start upload_robot_code

        # store sensor info
        self.robot_hardware = info['robot_hardware']
        self.has_ultra_sound = info['has_ultra_sound']
        self.has_camera = info['has_camera']
        self.has_bt_rssi = info['has_bt_rssi']
        self.has_uwb_module = info['has_uwb_module']
        self.has_accelerometer = info['has_accelerometer']
        self.has_speed = info['has_speed']
        self.has_ir_front = info['has_ir_front']
        self.has_ir_line = info['has_ir_line']
        self.has_led_stripe = info['has_led_stripe']
        # we use the robot's info to update our bias, we also update
        # the slider
        connection_param.left_right_bias = info['left_right_bias']
        local_param.shell_color = info['shell_color']

        logger.info(' [checkhand info: ultra-sound=%i, camera=%i, '
                    'bluetooth RRSI=%i, speed=%i, ir line=%i, '
                    'ir front=%i,]'
                    % (self.has_ultra_sound, self.has_camera,
                       self.has_bt_rssi, self.has_speed,
                       self.has_ir_front, self.has_ir_line))

        # update robot time
        current_time = time.asctime()
        self._send_object('set time', current_time)
        logger.debug("Robot time updated to {}".format(current_time))

    def check_hand_robot(self):
        if self.robot is not None:
            self.robot.blink_leds(
                period=1.5, color=self.robot.color, on=.5
            )

        # Signal threads that we are ready to establish a connection
        self.connection_status = ConnectionStatus.Ready

        # Start (or reuse) listening sockets and threads
        # Wi-Fi
        if self.wifi_connection_thread is None or \
                not self.wifi_connection_thread.is_alive() or \
                self.wifi_socket is None:
            # This is why it is important to set wifi_socket to None after
            # closing it! It is the only way to check if the socket is open
            # or not.
            if self.wifi_socket is not None:
                self.wifi_socket.close()

            self.wifi_socket = socket.socket(socket.AF_INET,
                                             socket.SOCK_STREAM)
            self.wifi_socket.setsockopt(socket.SOL_SOCKET,
                                        socket.SO_REUSEADDR, 1)
            self.wifi_socket.bind(('', _wifi_port))
            self.wifi_socket.settimeout(None)  # Wait indefinitely
            self.wifi_socket.listen(1)

            self.wifi_connection_thread = threading.Thread(
                target=self._wait_for_connection,
                args=(ConnectionStatus.Wifi, self.wifi_socket),
                name="wifi"
            )
            self.wifi_connection_thread.start()

        # Bluetooth
        if self.bt_connection_thread is None or \
                not self.bt_connection_thread.is_alive() or \
                self.bt_socket is None:
            # This is why it is important to set bt_socket to None after
            # closing it! It is the only way to check if the socket is open
            # or not.
            if self.bt_socket is not None:
                self.bt_socket.close()

            b_mac = Utils_bluetooth_for_robot.get_b_mac()
            self.bt_socket = socket.socket(
                socket.AF_BLUETOOTH,
                socket.SOCK_STREAM,
                socket.BTPROTO_RFCOMM
            )
            self.bt_socket.setsockopt(socket.SOL_SOCKET,
                                      socket.SO_REUSEADDR, 1)
            self.bt_socket.bind((b_mac, _bt_port))
            self.bt_socket.settimeout(None)  # Wait indefinitely
            self.bt_socket.listen(1)

            self.bt_connection_thread = threading.Thread(
                target=self._wait_for_connection,
                args=(ConnectionStatus.Bluetooth, self.bt_socket),
                name="bluetooth"
            )
            self.bt_connection_thread.start()

        # Wait until a connection is established
        while self.connection_status == ConnectionStatus.Ready:
            time.sleep(0.1)

        # Do not close listening sockets while corresponding threads are alive
        if self.connection_status == ConnectionStatus.Wifi:
            self.wifi_connection_thread.join()
            self.wifi_socket.close()
            self.wifi_socket = None
        elif self.connection_status == ConnectionStatus.Bluetooth:
            self.bt_connection_thread.join()
            self.bt_socket.close()
            self.bt_socket = None
        else:
            logger.warning("Failed to establish connection")
            return

        self.send_file = self.connection.makefile('wb')
        self.connected = True
        logger.info(' [accepted connection]')

        # send info
        info = {
            'robot_hardware': self.robot.hardware,
            'has_ultra_sound': self.robot.has_ultra_sound,
            'has_camera': self.robot.has_camera,
            'has_bt_rssi': self.robot.has_bt_rssi,
            'has_uwb_module': self.robot.has_uwb_module,
            'has_accelerometer': self.robot.has_accelerometer,
            'has_speed': self.robot.has_speed,
            'has_ir_line': self.robot.has_ir_line,
            'has_ir_front': self.robot.has_ir_front,
            'has_led_stripe': self.robot.has_led_stripe,
            'robot_code_version': robot_code_version,
            'left_right_bias': self.robot.left_right_bias,
            'shell_color': self.robot.color_name,
        }
        self._send_object('checkhand info', info)
        logger.info(' [sent checkhand info]')

        # read specific commands
        self.connection.settimeout(TIMEOUT_ROBOT_DIALOG)  # Set a long
        # timeout because if the robot code version is outdated, user has a
        # dialog to read and check, then we need to wait for robot code
        # complete upload.
        name, value = self._read_object(timeout=TIMEOUT_ROBOT_DIALOG)
        try:
            if name == 'wait for restart':
                # this happens if the computer detected that the robot
                # code version did not match it own robot_code_version
                # number, and therefore uploaded new code -> restart the
                # program
                logger.info(
                    'Received "wait for restart" while checking hand!'
                )
                name, value = self._read_object(timeout=TIMEOUT_ROBOT_DIALOG)
                if name != 'restart program':
                    raise RuntimeError('expected "restart program" '
                                       'command, received "%s"' % name)
                self.restart_robot_program()
            elif name == 'set time':
                os.system("sudo date -s '{}'".format(value))
            elif name == 'disconnect':
                # this was only a checkup connection, cancel connection
                # the cleanup will occur in the handling of this error in
                # self.check_hand
                raise ConnectionCheckOK
            else:
                raise ValueError(
                    'Expected "set time" command, received "%s".' % name)
        finally:
            if self.connection is not None:
                self.connection.settimeout(TIMEOUT_ROBOT)  # Set
                # reasonable timeout for future reads. We need it to correctly
                # detect disconnections.

    def restart_robot_program(self):
        if self.is_PC:
            logger.warning("PC received command 'restart program'!")
        else:
            logger.info('Received "restart program"')
            logger.info('Close channel')

            # Full cleanup
            self.cleanup()
            try:
                self.robot.cleanup()
            except Exception as e:
                logger.exception(e)
            # self.cleanup does not close connection sockets while the
            # corresponding thread is alive. But at this point, we need to
            # free the used addresses.
            if self.wifi_socket is not None:
                self.wifi_socket.close()
                self.wifi_socket = None
            if self.bt_socket is not None:
                self.bt_socket.close()
                self.bt_socket = None

            # TODO: check that the restart actually works!
            logger.info('Start new program')
            import subprocess
            command = ['sudo', 'python3', '/home/pi/AlphaI/Main_Client.py']
            subprocess.Popen(command, start_new_session=True)
            logger.info('Quit!')
            raise SystemExit

    def check_hand(self, check_only=False, b_mac=None):
        with self._lock:
            logger.debug('channel - lock acquired by check_hand')

            # connect and handle connection errors
            try:
                if not check_only:
                    logger.info('Hand checking ({} side)...'.format(self.side))
                # hand checking: it is actually the PC that tries to connect the
                # robot, because only the robot has a fix IP address...
                if self.is_PC:
                    info = self.check_hand_PC(check_only=check_only, b_mac=b_mac)
                    if check_only:
                        self.cleanup(check_only)
                        return info
                else:
                    self.check_hand_robot()

            except Exception as err:
                # connection failed, cleanup sockets before raising error again
                self.cleanup(check_only)
                if check_only:
                    # we are on PC side and attempted to connect to a
                    # non-responsive robot
                    if not isinstance(err, ConnectionCheckFail):
                        logger.debug("Check-only connection was started but not"
                                     " completed!")
                    return None
                raise err

            finally:
                if self.side == 'robot':
                    # turn off leds
                    self.robot.set_leds(0)
                logger.debug('channel - lock released by check_hand')

        logger.info('[channel connected]')

    # Perform robot code update from PC side
    def robot_code_update(self, remote_version):
        try:
            from Program.Utils import message_box, tr
        except ImportError:
            logger.warning("Failed to import Program.Utils")
            message_box = logger.info
            tr = lambda x: x

        if not connection_param.wifi_connection:
            message_box(
                (tr("Robot code version is currently %s.") + "\n" +
                 tr("It needs to be updated to %s.") + "\n" +
                 tr("Please connect to the robot using wifi to perform this "
                    "update.") + "\n" +
                 tr("Continuing without updating can cause bugs in "
                    "the AlphAI software.")) %
                (str(remote_version), str(robot_code_version))
            )
        else:
            message_box(
                (tr('Robot code version is currently %s.') + ' '
                 + tr('Version %s will be uploaded.'))
                % (str(remote_version), str(robot_code_version))
            )
            with show_wait_cursor():
                logger.info('Tell robot program to wait for restart signal')
                self._send_object('wait for restart',
                                  remote_version=remote_version)

                logger.info('Upload up-to-date robot code')
                Upload_Robot_Code.upload_robot_code()

                robot_auto_restart = (
                        not isinstance(remote_version, float)
                        or remote_version >= 2.1
                )
                if robot_auto_restart:
                    logger.info('Restart robot program')
                    self._send_object('restart program',
                                      remote_version=remote_version)
            # Raise an error because we cannot reconnect immediately.
            # This error is handled in Errors.py and will advise user to
            # reconnect once the robot has restarted.
            raise RobotCodeUploadedError(robot_auto_restart)

    def disconnect(self):
        if self.connected:
            try:
                self.send_command('disconnect')
            except Exception as err:
                logger.debug(err)
                display_error_traceback(err, developer_only=True)
        self.cleanup()

    def cleanup(self, check_only=False):
        if not check_only:
            logger.info(
                'Cleaning up communication channel ({} side)'.format(self.side)
            )
        self.connection_status = ConnectionStatus.NotConnected
        self.connected = False
        # Do not close wifi/Bluetooth sockets while corresponding threads
        # are running -> threads will be reused
        if self.wifi_socket is not None \
                and (self.wifi_connection_thread is None
                     or not self.wifi_connection_thread.is_alive()):
            self.wifi_socket.close()
            self.wifi_socket = None  # Signals that it has been closed
        if self.bt_socket is not None \
                and (self.bt_connection_thread is None
                     or not self.bt_connection_thread.is_alive()):
            self.bt_socket.close()
            self.bt_socket = None  # Signals that it has been closed
        if self.send_file and not self.send_file.closed:
            try:
                self.send_file.close()
            except Exception as e:
                # not clear why we sometimes get an error here
                logger.debug(e)
            finally:
                self.send_file = None
        if self.connection is not None:
            self.connection.close()
            self.connection = None
        if not self.is_PC:
            try:
                self.robot.stop()
            except:
                # can happen if self.robot.cleanup() has been called
                pass

    def is_closed(self):
        return not self.connected or self.send_file is None or \
                        self.send_file.closed

    def _send_bytes(self, x: bytes):
        # uncomment line below for debugging
        # print("sending", x)
        self.send_file.write(x)

    def _read_bytes(self, n: int, timeout=TIMEOUT_DEFAULT) -> bytes:
        read_buffer = b""
        self.connection.settimeout(timeout)
        starting_time = time.perf_counter()
        while len(read_buffer) < n:
            if time.perf_counter() - starting_time > timeout:
                raise socket.timeout
            read_buffer += self.connection.recv(min(4096, n-len(read_buffer)))
        return read_buffer

    def _send_object(self, name: str, value=None,
                     remote_version=robot_code_version):
        with self._lock:
            logger.debug('channel - lock acquired by send_object (%s)' % name)

            # send and handle connection errors
            try:
                # name
                x = str.encode(name)
                self._send_bytes(struct.pack('<H', len(x)))
                self._send_bytes(x)

                # value
                x = pickle.dumps(value)
                self._send_bytes(struct.pack('<L', len(x)))
                self._send_bytes(x)

                self.send_file.flush()
                return

            except AttributeError:
                self.disconnect()

            finally:
                if name == 'disconnect':
                    self.connected = False
                logger.debug('channel - lock released by send_object (%s)' %
                             name)

    def _read_object(self, wait=True, timeout=TIMEOUT_DEFAULT):
        with self._lock:
            logger.debug('channel - lock acquired by read_object')
            # if we are aware that we are not connected
            if not self.connected:
                logger.debug(
                    'attempted to read while not being connected, return None'
                )
                logger.debug('channel - lock released by read_object (None)')
                return '', None

            # read and handle connection errors
            try:
                try:
                    # reading name
                    if not wait:
                        two_bytes = self._read_bytes(2,
                                                     timeout=TIMEOUT_PC_CHECK)
                    else:
                        two_bytes = self._read_bytes(2, timeout=timeout)
                    n, = struct.unpack('<H', two_bytes)
                    name = self._read_bytes(n, timeout=timeout).decode()

                    # reading value
                    four_bytes = self._read_bytes(4, timeout=timeout)
                    n, = struct.unpack('<L', four_bytes)
                    data = self._read_bytes(n, timeout=timeout)
                    value = pickle.loads(data)

                    self.first_no_data_time = None
                    return name, value

                except Exception as err:
                    # no waiting: timeout just means that there is no data yet,
                    # return as long as the first read attempt was no longer
                    # than TIMEOUT_PC_READ before
                    if not wait and isinstance(err, socket.timeout):
                        if self.first_no_data_time is None:
                            self.first_no_data_time = time.perf_counter()
                        if time.perf_counter() < self.first_no_data_time + \
                                TIMEOUT_PC_READ:
                            return None

                    # otherwise there is a connection error
                    raise err
            except ConnectionAbortedError:
                self.disconnect()
            except socket.timeout:
                self.disconnect()
            except Exception as err:
                logger.debug(err)
            finally:
                try:
                    msg = 'received %s' % name
                except NameError:
                    msg = 'read failed'
                logger.debug('channel - lock released by read_object (%s)' %
                             msg)

    def scan_bluetooth_devices(self):
        self.send_command('bluetooth scan', None)
        self.connection.settimeout(TIMEOUT_BT_SCAN)
        name, addr_list = self.read_sensor()
        self.connection.settimeout(TIMEOUT_PC_READ)
        return addr_list.tolist()


if __name__ == "__main__":

    import numpy as np

    # # prepare display
    # plt.ion()
    # f = plt.figure(figsize=(15, 8))
    # f.show()
    # ax = plt.axes()
    # ax.set_ylim((-2.5, 6.5))
    # points = np.zeros((1, 4))
    # lines = ax.plot(points[:, 1:4])
    # averages = np.zeros((1, 4))
    # avg_lines = ax.plot(averages[:, 1:4], '.-')
    # t_max = 5
    # ax.set_xlim((0, t_max))
    #
    # # Stop-Start button
    # from PyQt5 import QtWidgets
    # button = QtWidgets.QPushButton()
    # button.setText('START-STOP')
    # button.setCheckable(True)
    robot_move = False
    # def start_stop():
    #     global robot_move
    #     robot_move = button.isChecked()
    # button.toggled.connect(start_stop)
    # button.show()

    channel = CommunicationChannel('PC', robot_ip='10.0.0.1')


    def test_send():
        logger.debug('sending')
        global robot_move
        if robot_move:
            channel.send_command('motor', [40., 40.])
        else:
            channel.send_command('motor', [0., 0.])
        logger.debug('sending done')


    def test_read(wait=True):
        value = channel.read_sensor(wait)
        try:
            name, value = value
            return value.__dict__
        except:
            logger.debug('<', value, '>')


    # channel.send_command('use UWB', False)
    # channel.send_command('use UWB', True)
    test_send()
    test_read()

    # colors = np.repeat(np.array([[1., 0., 0.]]), 4, axis=0)
    colors = [[0., .5, 0.]] * 4
    channel.send_command('leds', colors)

    # IRP = IRport
    # print("activate IR")
    # channel.send_command('use IR', int(IRP.LineExtrem))
    # print("done activate IR")
    #
    # test_send()
    # test_read()
    # test_send()
    # test_read()
    #
    # _print('use UWB', False)
    #
    # # _print('bluetooth scan')
    # # print(channel.scan_bluetooth_devices())
    #
    # _print('use UWB', True)
    # # channel.send_command('use UWB', True)
    #
    #
    # # test_send()
    # # test_read()
    # # Tell robot to use accelerometer
    # channel.send_command('use accelerometer', -1)
    #
    # j = 0
    # while True:
    #     test_send()
    #     val = test_read()
    #     vals = val.get("accelerometer", "none")
    #     vals = np.array(vals)
    #
    #     # concatenate all points
    #     print(points.shape, vals.shape)
    #     if len(vals) > 0:
    #         points = np.vstack((points, vals))
    #     last_time = points[-1, 0]
    #     ok_time = (last_time - points[:, 0]) <= t_max
    #     points = points[ok_time, :]
    #
    #     # display points
    #     times = points[:, 0] - points[0, 0]
    #     for i in range(3):
    #         lines[i].set_xdata(times)
    #         lines[i].set_ydata(points[:, 1+i] + 2*i)
    #
    #     # compute and concatenate average values
    #     avg_vals = np.mean(vals, axis=0, keepdims=True)
    #     # avg_vals[0, 1:] = averages[-1, 1:] + avg_vals[0, 1:]
    #     averages = np.vstack((averages, avg_vals))
    #     ok_time = (last_time - averages[:, 0]) <= t_max
    #     averages = averages[ok_time, :]
    #     print(averages.shape)
    #
    #     # display average values
    #     times = averages[:, 0] - points[0, 0]
    #     for i in range(3):
    #         avg_lines[i].set_xdata(times)
    #         x = averages[:, 1 + i]
    #         x = x[0] + 4 * (x - x[0])
    #         avg_lines[i].set_ydata(x + 2 * i + .5)
    #
    #     plt.pause(0.001)

    # print('restart robot program')
    # channel.send_command('restart program')
    #
    # print('wait 2s')
    # time.sleep(2)
    #
    # print('attempt to reconnect!')
    # test_send()
    # test_read()

    # channel.send_command('use BT rssi', False)
    # test_send()
    # test_read()
    #
    # test_send()
    # test_read(wait=False)
    # test_read()
    #
    # # print('erroneous second reading in a row')
    # # test_read()
    # #
    # # test_send()
    # # test_read()
    # #
    # # test_send()
    # #
    # # print('erroneous second send in a row')
    # # test_send()
    # # test_read()
    # #
    # # test_send()
    # # test_read()
    #
    # print('activitating BT rssi scan')
    # # channel.send_command('use BT rssi', "14:A1:D0:DD:7F:7A")
    # channel.send_command('use BT rssi', "C4:09:1D:79:91:AF")
    # # channel.send_command('use BT rssi', "27:74:6A:BD:53:C5")
    # test_send()
    # test_read()
    # test_send()
    # test_read()
    # test_send()
    # test_read()
