# https://github.com/pimoroni/rpi_ws281x-python/blob/master/examples/lowlevel.py

import _rpi_ws281x as ws
import numpy as np
from RobotLogger import robot_logger as logger


# LED configuration.
LED_CHANNEL = 0
LED_COUNT = 4         # How many LEDs to light.
LED_FREQ_HZ = 800000  # Frequency of the LED signal. Should be 400 or 800khz
LED_DMA_NUM = 10      # DMA self.channel to use, can be 0-14.
LED_GPIO = 18         # GPIO connected to the LED line.  Must support PWM!
LED_BRIGHTNESS = 255  # Set to 0 for darkest and 255 for brightest
LED_INVERT = 0        # Set to 1 to invert the LED signal, good if using NPN

DOT_COLORS = [0x200000,  # green
              0x201000,  # yellow
              0x202000,  # orange
              0x002000,  # red
              0x002020,  # purple
              0x000020,  # blue
              0x100010,  # light-blue
              0x200010,  # white
              0x000000]  # off

# TODO: use Flags to be able simply to set a sub-set of leds
STRIPE = ('right_1',
          'right_2',
          'left_1',
          'left_2')


def colors_by_name(color_name: str):
    colors = {
        "white": [[1., 1., 1.]] * 4,
        "black": [
            [0., 0., 1.], [1., 1., 1.], [1., 1., 1.], [1., 0., 0.]
        ],
        "red": [[1., 0., 0.]] * 4,
        "blue": [[0., 0., 1.]] * 4,
        "green": [[0., 1., 0.]] * 4,
        "magenta": [[1., 0., 1.]] * 4,
        "yellow": [[1., 1., 0.]] * 4,
        "orange": [[1., .5, 0.]] * 4,
    }.get(color_name, 1)
    return colors


class LedStripe:
    def __init__(self):
        # self.led_stripe_thread = None
        self.leds = ws.new_ws2811_t()

        # Initialize all channels to off
        for chan_num in range(2):
            self.channel = ws.ws2811_channel_get(self.leds, chan_num)
            ws.ws2811_channel_t_count_set(self.channel, 0)
            ws.ws2811_channel_t_gpionum_set(self.channel, 0)
            ws.ws2811_channel_t_invert_set(self.channel, 0)
            ws.ws2811_channel_t_brightness_set(self.channel, 0)

        self.channel = ws.ws2811_channel_get(self.leds, LED_CHANNEL)

        ws.ws2811_channel_t_count_set(self.channel, LED_COUNT)
        ws.ws2811_channel_t_gpionum_set(self.channel, LED_GPIO)
        ws.ws2811_channel_t_invert_set(self.channel, LED_INVERT)
        ws.ws2811_channel_t_brightness_set(self.channel, LED_BRIGHTNESS)

        self.brightness = LED_BRIGHTNESS

        ws.ws2811_t_freq_set(self.leds, LED_FREQ_HZ)
        ws.ws2811_t_dmanum_set(self.leds, LED_DMA_NUM)

        # Initialize library with LED configuration.
        resp = ws.ws2811_init(self.leds)
        if resp != ws.WS2811_SUCCESS:
            message = ws.ws2811_get_return_t_str(resp)
            raise RuntimeError(
                'ws2811_init failed with code {0} ({1})'.format(resp, message))

    def set_led_colors(self, colors):
        # colors is an 4x3 array of float between 0 and 1
        # Warning: input colors are RGB values, but value sent to
        # ws2811_led_set should be ordered as green-red-blue!!
        if ws is None:
            logger.debug("ws is None")
            return
        for i, x in enumerate(colors):
            if len(x) != 3:
                raise ValueError('Colors must be 3-element arrays of floats')
            x = [int(max(0., min(1., xk)) * 255) for xk in x]
            color = (x[1] << 16) + (x[0] << 8) + x[2]
            ws.ws2811_led_set(self.channel, 3 - i, color)
        resp = ws.ws2811_render(self.leds)
        if resp != ws.WS2811_SUCCESS:
            message = ws.ws2811_get_return_t_str(resp)
            raise RuntimeError(
                'LED stripe: ws2811_render failed with code %i (%s)'
                % (resp, message))

    def switch_off(self):
        self.set_led_colors(np.zeros((4, 3)))

    # this is used to avoid memory leak (see github link above)
    def terminate_led_communication(self):
        try:
            # Ensure ws2811_fini is called before the program quits.
            ws.ws2811_fini(self.leds)
            # Example of calling delete function to clean up structure memory.
            # Isn't strictly necessary at the end of the program execution
            # here, but is good practice.
            ws.delete_ws2811_t(self.leds)
        finally:
            pass

    def __del__(self):
        self.terminate_led_communication()
