import RobotInfo
import os
from socket import timeout
from RobotLogger import robot_logger as logger
import time
import threading
import subprocess
from Environment.AlphAI import Communication
import AlphaBot

try:
    from LedStripe import LedStripe
    from BootingLeds import BootingLeds as WaitingLed

    time.sleep(5)  # LedStripe initialized before about 4 seconds after
    # startup will cease to function at this time. The reason is
    # currently unknown.
    led_stripe = LedStripe()
    color = RobotInfo.get_robot_info('color')
    waiting_led = WaitingLed(led_stripe, color)

except ImportError:
    class WaitingLed:
        def __enter__(self):
            return self

        def start(self):
            pass

        def stop(self):
            pass

        def done(self):
            pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            return True


    led_stripe = LedStripe = None
    waiting_led = WaitingLed()

waiting_led.start()


class Program:

    def __init__(self):
        # Robot object
        self.alpha_bot = AlphaBot.AlphaBot(led_stripe=led_stripe)

        self.alpha_bot.s_button.when_released = self.switch_program

        self.channel = Communication.CommunicationChannel(
            'robot', robot=self.alpha_bot
        )

        self.report_wifi()

        # Bluetooth configuration: after a delay to make sure Raspberry OS
        # has finished its own Bluetooth configuration
        time.sleep(5)
        import Utils_bluetooth_for_robot
        logger.debug("Attempting to initialize Bluetooth")
        Utils_bluetooth_for_robot.init_bluetooth()

        # When the Wi-fi and Bluetooth connections are ready, stop waiting
        # leds and do a small jiggle
        waiting_led.done()
        self.alpha_bot.jiggle(blocking=True)

        self.main_loop()

    def report_wifi(self):
        # Report wifi status
        try:
            wifi_ok = False
            while not wifi_ok:
                command = ["/sbin/iwconfig"]

                output = subprocess.check_output(
                    command, stderr=subprocess.STDOUT
                ).decode()
                if 'Mode:Master' in output:
                    logger.debug("Wi-Fi network is ready.")
                    wifi_ok = True
                else:
                    time.sleep(.5)

        except Exception as err:
            logger.error('could not check wifi')
            logger.exception(err)

    def main_loop(self):
        # Outer loop: restart from zero if some unhandled error occurs
        while True:
            try:
                # Open communication channel
                while True:  # we need to be able to restart the channel with a
                    # different protocol in case of timeout
                    self.channel.check_hand()
                    # the connection is stored in the Ab object
                    #  if the link is ok, we break and start
                    if self.channel.connected:
                        break
                self.alpha_bot.jiggle(3)

                # Main loop: get and execute commands from PC
                while not self.channel.is_closed():
                    _command = self.channel.read_command()
                    if _command is not None:
                        name, value = _command
                    else:
                        logger.warning("Missing data, onto the next command")
                        continue
                    logger.debug('received {} {}'.format(name, value))

                    if name == 'step_sensor_ahead':
                        # move (and buzz) for a fixed duration; this will
                        # not wait for the duration to finish but rather
                        # start a thread that will stop the robot at the end
                        # of the duration
                        left, right, buzz, duration, read_sensor_ahead = value
                        self.alpha_bot.set_motor_and_buzzer(
                            left, right, buzz, duration
                        )

                        # read data a fixed delay BEFORE the duration has
                        # finished note that read_sensor_ahead is a negative
                        # value
                        if duration + read_sensor_ahead > 0:
                            time.sleep(duration + read_sensor_ahead)
                        data = self.alpha_bot.get_sensor_data()
                        self.channel.send_sensor('all sensors', data)

                        # note that we usually will reach here before the
                        # duration has finished, so a new command will also
                        # be received before this; the new command will call
                        # again Ab.set_motor_and_buzzer which will first
                        # wait for the previous duration to finish and then
                        # execute the new command

                    elif name == 'set motors':
                        left, right, duration = value
                        self.alpha_bot.stop()
                        self.alpha_bot.set_motor(left, right)
                        if duration is not None:  # use a timer to prevent
                            # timeout when asking too long durations
                            t = threading.Timer(duration, self.alpha_bot.stop)
                            t.start()

                    elif name == 'step':
                        # move the motors, but only during a fixed amount of
                        # time
                        buzz = False
                        pause = None
                        if len(value) == 3:
                            left, right, duration = value
                        elif len(value) == 4:
                            left, right, buzz, duration = value
                        else:
                            left, right, buzz, duration, pause = value
                        self.alpha_bot.set_motor(left, right)
                        self.alpha_bot.set_buzzer(buzz)
                        time.sleep(duration)
                        self.alpha_bot.stop()
                        if pause is not None:
                            time.sleep(pause)

                        # Send sensor data
                        data = self.alpha_bot.get_sensor_data()
                        self.channel.send_sensor('all sensors', data)

                    elif name == 'query sensor':
                        # Send sensor data
                        data = self.alpha_bot.get_sensor_data()
                        self.channel.send_sensor('all sensors', data)

                    elif name == 'disconnect':
                        # Disconnect: simply break from this loop and
                        # cleanup will occur in the "finally" block
                        self.alpha_bot.stop()  # don't forget to stop the robot
                        break

                    elif name == 'leds':
                        if self.alpha_bot.has_led_stripe:
                            self.alpha_bot.set_leds(value)

                    elif name == 'blink leds':
                        if self.alpha_bot.has_led_stripe:
                            self.alpha_bot.blink_leds(*value)

                    elif name == 'set default led':  # the leds will fade
                        # toward the new color
                        self.alpha_bot.set_leds_off_color(
                            value)  # default color is private
                        # we use the private methode

                    elif name == 'set robot color':  # the leds will fade
                        # toward the new color
                        self.alpha_bot.color = value

                    elif name == 'buzzer':
                        self.alpha_bot.set_buzzer(value)

                    elif name == 'stop':
                        # use this command to stop the robot without requesting
                        # an upcoming sensor data
                        self.alpha_bot.stop()

                    elif name == 'bluetooth scan':
                        bt_devices = self.alpha_bot.get_bt_devices()
                        self.channel.send_sensor('bt_devices', bt_devices)

                    elif name == 'use BT rssi':
                        if value:
                            logger.info(
                                'trying to get RSSI of bluetooth '
                                'device {}'.format(value)
                            )
                        else:
                            logger.info('not getting bluetooth RSSI value')
                        self.alpha_bot.use_BT_rssi(value)

                    elif name == 'use UWB':
                        logger.info(
                            "{} using UWB module".format(
                                ('' if value else 'not '))
                        )
                        self.alpha_bot.use_UWB(value)

                    elif name == 'use accelerometer':
                        if value:
                            logger.info(
                                'use accelerometer with read period {}'.format(
                                    value)
                            )
                        else:
                            logger.info('not using accelerometer')
                        self.alpha_bot.use_accelerometer(value)

                    elif name == 'use IR front':
                        # value is of type Communication.IRport
                        if value:
                            logger.info('use front infra-red')
                        else:
                            logger.info('not using front infra-red')
                        self.alpha_bot.use_IR_front(value)

                    elif name == 'use IR line':
                        # value is of type Communication.IRport
                        if value:
                            logger.info(
                                'read line-tracking infra-red sensors '
                                '{}'.format(value)
                            )
                        else:
                            logger.info('not using line-tracking infra-red')
                        self.alpha_bot.use_IR_line(value)

                    elif name == 'use camera':
                        if value:
                            logger.info(
                                'use camera with resolution {}'.format(value))
                        else:
                            logger.info('not using camera')
                        # value must be camera resolution (2-element vector)
                        # or False (or None)
                        self.alpha_bot.use_camera(value)

                    elif name == 'update robot info':
                        # the user manually change the bias, the robot save it
                        self.alpha_bot.update_left_right_bias(value)

                    elif name == 'set time':
                        os.system("sudo date -s '{}'".format(value))

                    elif name == 'restart program':
                        # Start a new program!, usually with updated code
                        self.channel.restart_robot_program()

                    elif name == "maintain connection":
                        # nothing to do, this command was only used to avoid
                        # reading timeout
                        pass

                    else:
                        logger.info('unknown command name %s!' % name)

            except KeyboardInterrupt:
                logger.info('Keyboard interrupted, terminating')
                break

            except timeout:
                logger.warning("Connection lost, cleanup")

            except Exception as err:
                logger.exception(err)

            finally:
                if self.channel is not None:
                    self.channel.cleanup()

    def switch_program(self):
        logger.info('Received "switch program"!')
        logger.info('Close channel')
        self.channel.cleanup()
        time.sleep(.1)

        logger.info('Start other program')
        import subprocess

        subprocess.Popen(
            ['sudo', 'python3', '/home/pi/AlphaI/ButtonClient.py',
             str(os.getpid()), '&']
        )

    def __del__(self):
        if self.channel is not None:
            self.channel.cleanup()
        self.alpha_bot.cleanup()


if __name__ == '__main__':
    Program()
