import time

import numpy as np
import os


class DummyLogger:
    @staticmethod
    def info(msg):
        print(msg)

try:
    from RobotLogger import robot_logger as logger
except FileNotFoundError:
    logger = DummyLogger()
except ModuleNotFoundError:
    logger = DummyLogger()


class KNN:
    _has_warned = False

    def __init__(self, robot, n_input, n_output, k, continuous, actions,
                 duration, size=1000, dtype=np.float32):
        self.robot = robot  # type:AlphaBot
        self.n_input = n_input
        self.n_output = n_output
        self.n_neighbour = k
        self.continuous = continuous
        self.available_actions = actions
        self.duration = duration
        self.buffer_size = size
        self.dtype = dtype
        self._init_model()

    def _init_model(self):
        if self.continuous:
            self._knn = self._continuous_knn
        else:
            self._knn = self._discrete_knn
        self._neighbour_index = []
        # ALGO_KNN has no meaning without an experience buffer,
        # set use_buffer parameter to True in case it is not the case yet
        logger.debug('init experience buffer')
        buffer_size = self.buffer_size

        # defining a circular array
        self.buffer_fill = 0
        self.buffer_pos = 0  # cursor position for next write action
        self.state_buffer = np.empty(
            (buffer_size, self.n_input), dtype=self.dtype
        )
        self.action_buffer = np.empty(
            (buffer_size, 2)
        )  # note that tf.one_hot currently accepts only uin8, int32 and int64
        self.state_buffer_auto = np.empty(
            (buffer_size, self.n_input), dtype=self.dtype
        )
        self.action_buffer_auto = np.empty(
            (buffer_size, 2)
        )  # note that tf.one_hot currently accepts only uin8, int32 and int64
        self.buffer_pos_auto = 0

    @property
    def k(self):
        return self.n_neighbour

    @k.setter
    def k(self, value):
        if not self.n_neighbour == value:
            logger.info("k has been updated to {}".format(value))
            self.n_neighbour = value

    def increase_k(self):
        self.k += 1

    def decrease_k(self):
        if self.k > 1:
            self.k -= 1

    @property
    def actions(self):
        return self.available_actions

    @actions.setter
    def actions(self, value):
        logger.info("actions have been updated to {}".format(value))
        self.available_actions = value

    def get_buffer(self, n_last=None):
        # obtain the experience buffer, or a subpart of it
        if n_last is None:
            n_last = self.buffer_fill
        if n_last <= self.buffer_pos:
            # read values only before self.buffer_pos -> use a slice to get
            # a view on the full array
            idx = slice(self.buffer_pos - n_last, self.buffer_pos)
        else:
            # need to read some values at the end of the current circular
            # arrays in addition to values before the cursor position
            # self.buffer_pos -> new arrays will be created
            n_end = n_last - self.buffer_pos
            current_size = len(self.state_buffer)
            idx = np.hstack(
                (np.arange(current_size - n_end, current_size),
                 np.arange(0, self.buffer_pos))
            )
        return dict(
            (
                ("action", self.action_buffer[idx]),
                ("state", self.state_buffer[idx]),
            )
        )

    def _resize_buffer(self):
        # change resize, learn funcs and n_input entries
        # think about whether we need n_shape
        # probs: right action, .., ..
        new_size = self.buffer_size
        current_size = len(self.state_buffer)
        if new_size < current_size:
            # create new buffer arrays of smaller size by reading at the
            # appropriate positions
            buffer = self.get_buffer(n_last=new_size)
            for key, value in buffer.items():
                setattr(self, key + '_buffer', value)
            self.buffer_fill = new_size
            self.buffer_pos = 0
        elif new_size > current_size:
            # we first make the arrays larger
            self.action_buffer = np.resize(self.action_buffer, (new_size, 2))
            self.state_buffer = np.resize(
                self.state_buffer, (new_size, self.n_input)
            )
            # if previous buffer was not filled yet, we are done, otherwise
            # we need to reposition values
            if self.buffer_fill == current_size:
                idx = np.hstack((
                    np.arange(self.buffer_pos, current_size),
                    np.arange(0, self.buffer_pos)
                ))
                self.action_buffer[:current_size] = self.action_buffer[idx]
                self.state_buffer[:current_size] = self.state_buffer[idx]
                self.buffer_pos = current_size

    def learn(self, action, state):
        self._update_model(action, state)

    def forget_last_second(self):
        if self.buffer_pos < int(1 / self.duration):
            self.buffer_pos = 0
            self.buffer_fill = 0
        else:
            self.buffer_pos -= int(1 / self.duration)
            self.buffer_fill -= int(1 / self.duration)
        logger.info(
            "removing last second of recording: buffer position is "
            "{}".format(self.buffer_pos))

    def _update_model(self, action, state):
        if action is None:
            # This happens when all connections are zero: no action is
            # chosen, so no learning can occur
            return 0

        # Add last experience to buffer
        if self.buffer_size != self.state_buffer.shape[0]:
            # buffer size has been set to a new value: resize buffer
            self._resize_buffer()
        self.action_buffer[self.buffer_pos, :] = action
        self.state_buffer[self.buffer_pos] = state
        self.buffer_pos = (self.buffer_pos + 1) % self.buffer_size
        if self.buffer_pos == self.buffer_size - 1 and not self._has_warned:
            self.robot.play_jingle()
            self._has_warned = True  # buzz only once
        if self.buffer_fill < self.buffer_size:
            self.buffer_fill += 1
        logger.info(
            'experience buffer has {} elements'.format(self.buffer_fill))

    def _continuous_knn(self, state):
        # state is a 2D array: n_batch*xy
        n_batch = len(state)

        # training data
        buffer = self.get_buffer()
        train_states = buffer["state"]  # n_train*xy
        n_train = len(train_states)

        # no training data yet?
        k = min(n_train, self.k)
        votes = np.zeros(
            (n_batch, self.n_output), dtype=int
        )  # n_batch*n_action
        idx_neighbors = np.zeros((n_batch, k), dtype=int)
        if n_train == 0:
            return votes, None, idx_neighbors

        # compute and sort distances to all points in training data
        state = state.astype(self.dtype)
        state2, train_states2 = np.broadcast_arrays(
            state[:, np.newaxis, :],
            train_states[np.newaxis, :, :]
        )  # n_batch*n_train*n_state
        if self.dtype == np.uint8:
            # caution with subtracting uint8 values! we compute the absolute
            # difference as differenceImmageV6 on
            # https://stackoverflow.com/a/57771547)
            diff_mod256 = state2 - train_states2
            diff_sign_mod256 = np.uint8(state2 < train_states2) * 254 + 1
            diff_absolute = diff_mod256 * diff_sign_mod256
        elif self.dtype == np.float32:
            diff_absolute = state2 - train_states2
        else:
            raise RuntimeError('Invalid case', self.dtype)
        distance_list = np.linalg.norm(diff_absolute, axis=2)  # n_batch*n_train

        # k nearest neighbors sorted
        for i in range(k):
            idx_neighbors[:, i] = np.argmin(distance_list, axis=1)
            distance_list[range(n_batch), idx_neighbors[:, i]] = np.inf

        # take average action among k neighbors
        actions = buffer["action"]
        decisions = actions[idx_neighbors, :]  # n_batch*k*2
        decisions = np.mean(decisions, axis=1)  # n_bach*2
        return None, decisions, idx_neighbors

    def _discrete_knn(self, state):
        # state is a 2D array: n_batch*xy
        n_batch = len(state)

        # training data
        buffer = self.get_buffer()
        train_states = buffer["state"]  # n_train*xy
        n_train = len(train_states)

        # no training data yet?
        k = min(n_train, self.k)
        votes = np.zeros(
            (n_batch, self.n_output), dtype=int
        )  # n_batch*n_action
        idx_neighbors = np.zeros((n_batch, k), dtype=int)
        if n_train == 0:
            return votes, None, idx_neighbors

        # compute and sort distances to all points in training data
        vectors = state[:, np.newaxis, :] - \
                  train_states[np.newaxis, :, :]  # n_batch*n_train*xy
        distance_list = np.linalg.norm(vectors, axis=2)  # n_batch*n_train

        # k nearest neighbors sorted
        for i in range(k):
            idx_neighbors[:, i] = np.argmin(distance_list, axis=1)
            distance_list[range(n_batch), idx_neighbors[:, i]] = np.inf

        # votes for possible actions according to the k nearest neighbors
        actions = buffer["action"]
        for i in range(k):
            votes[range(n_batch), actions[idx_neighbors[:, i]]] += 1

        # when there is some equality between several actions, we want to
        # ignore the points that are the farther:
        # decrease number of neighbors as long as several actions reach the
        # maximal number of votes (this is tricky!)
        votes_trim = votes.copy()
        max_vote = votes_trim.max(axis=1, keepdims=True)
        several_candidates = (
                np.sum(votes_trim == max_vote, axis=1) > 1
        )  # n_batch
        idx, = np.where(several_candidates)
        i = k
        while len(idx):
            i -= 1
            actions_to_remove = actions[idx_neighbors[idx, i]]
            votes_trim[idx, actions_to_remove] -= 1
            several_candidates = np.logical_and(
                several_candidates,
                np.sum(votes_trim == max_vote, axis=1) > 1
            )  # n_batch
            idx, = np.where(several_candidates)

        # take decision on this "trimmed" votes
        decisions = votes_trim.argmax(axis=1)

        # return the trimmed votes, but all neighbors (included those that
        # were ignored during trimming)
        return votes_trim, decisions, idx_neighbors

    def get_decision_and_neighbors(self, state):
        # create a batch of only 1 value to call _knn method
        # tic = time.perf_counter()
        _, decision, idx_neighbors = self._knn(state[np.newaxis, :])
        # logger.info("Decision taken in {:.3f} ms".format(
        #     (time.perf_counter() - tic) * 1e3))
        # extract decision
        decision = None if decision is None else decision[0]
        # extract neighbors: note that there can be less neighbors than k in
        # the case where there is some equality in votes
        idx_neighbors = idx_neighbors[0]  # list of lists -> list
        idx_neighbors = idx_neighbors[idx_neighbors != -1]  # when less than k
        # neighbors were used, voids in array were filled with value -1
        return decision, idx_neighbors

    def get_and_memorize_decision(self, state):
        decision, _ = self.get_decision_and_neighbors(state)
        if decision is not None:
            self.state_buffer_auto[self.buffer_pos_auto] = state
            self.action_buffer_auto[self.buffer_pos_auto] = decision
            self.buffer_pos_auto = (self.buffer_pos_auto + 1) % self.buffer_size
        return decision

    def save_buffers_to_file(self):
        logger.info("Saving data buffer")
        np.save('state.npy', self.state_buffer[:self.buffer_fill])
        np.save('action.npy', self.action_buffer[:self.buffer_fill])
        np.save('state_auto.npy',
                self.state_buffer_auto[:self.buffer_pos_auto])
        np.save('action_auto.npy',
                self.action_buffer_auto[:self.buffer_pos_auto])

    def load_buffers_from_file(self, folder=''):
        def get_buffer(file, shape):
            f = os.path.join(folder, file)
            if os.path.isfile(f):
                buffer = np.load(f)
                length, width = buffer.shape
                return np.vstack(
                    (buffer, np.empty(
                        (self.buffer_size - length, width), dtype=self.dtype))
                ), length
            else:
                return np.empty(shape, dtype=self.dtype), 0

        self.state_buffer, _ = get_buffer(
            'state.npy', (self.buffer_size, self.n_input)
        )
        self.action_buffer, self.buffer_fill = get_buffer(
            'action.npy', (self.buffer_size, 2)
        )
        self.buffer_pos = self.buffer_fill
        logger.info(
            'training buffer has {} elements'.format(self.buffer_fill))

    def _reset_model(self):
        self._init_model()
