import threading
from time import sleep, time
import numpy as np
from collections import deque


class ThreadUnit(threading.Thread):

    def __init__(self, read_func, read_period=1e-3, debug=False, n_repeat=0):
        super().__init__()

        self.read_period = read_period
        self._do_stop = True
        self.last_measurements = deque()
        self.read_func = read_func
        self.n_sample_max = 5000
        self._n_repeat_remain = n_repeat
        self.debug = debug

    def run(self):
        # read sample continuously,
        # - if read_period is > 0, read at fixed times
        # - if read_period is -1, read continuously

        # reinit memory
        self.last_measurements = deque()

        # init ticks
        self._do_stop = False
        read_period = self.read_period if self.read_period > 0 else 0 # when self.read_period is -1 for continuous reading, read_period is 0
        if read_period:
            time_last_read = time() / self.read_period # time unit: self.read_period

        # measurement loop
        while not self._do_stop:
            if read_period:
                t = time() / self.read_period
                time_since_last_read = t - time_last_read
            if read_period and time_since_last_read > 1:
                # we missed some sample
                # print('missed sample')
                # tick = time()
                # self.last_measurements.append((np.NaN, np.NaN, np.NaN))
                # print('store fake in', (time()-tick)*1e3, 'ms')
                pass
            else:
                # wait for next tick
                if read_period:
                    # print('sleep', 1 - time_since_last_read)
                    sleep((1 - time_since_last_read) * self.read_period) # time unit: second
                # read
                # tick = time()
                measure = self.read_func()
                if measure is not None:
                    self.last_measurements.append((time(), *measure))
                    if self.debug:
                        print(measure)
                # print('read + store in', (time()-tick)*1e3, 'ms')
            if read_period:
                time_last_read += 1
            # do not keep more than n_sample_max samples
            while len(self.last_measurements) > self.n_sample_max:
                # print('remove sample')
                self.last_measurements.popleft()
            # max number of repeats?
            if self._n_repeat_remain:
                self._n_repeat_remain -= 1
                if self._n_repeat_remain == 0:
                    self._do_stop = True

    def stop(self, blocking=True):
        self._do_stop = True
        # wait for thread to terminate
        if blocking:
            self.join()

    def collect_measurements(self):
        n = len(self.last_measurements)
        t = time()
        ret = np.array([self.last_measurements.popleft() for _ in range(n)])
        # print('collected %i measurements in %.2f ms' % (n, (time()-t)*1e3))
        return ret
