import sys
import time
import subprocess
from configparser import ConfigParser
from RobotLogger import robot_logger as logger

try:
    import pexpect
    old_version = False
except ImportError:
    pexpect = None
    old_version = True


def get_ssid():
    with open("/etc/hostapd/hostapd.conf", "r") as hostapd_file:
        for line in hostapd_file:
            if line.startswith("ssid="):
                return line[5:].strip()
    logger.error("hostapd.conf file does not contain a ssid line")
    return 0


def init_b_name():  # easier with btmgmt name ALPHAI-XXXXXX?
    open_blue = subprocess.Popen(
        ["bluetoothctl"],
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        stdin=subprocess.PIPE
    )
    out, err = open_blue.communicate(
        ("system-alias " + get_ssid()).encode('ascii')
    )


def wait_for_bt_init():
    command = ["sudo", "hciconfig", "hci0"]  # could use btmgmt?
    while True:
        try:
            subprocess.run(command, stderr=subprocess.STDOUT, check=True)
            break
        except subprocess.CalledProcessError:
            time.sleep(0.5)
    logger.debug("Bluetooth device hci0 detected")


def set_bluetooth_config():
    try:
        if not old_version:
            # Check bluetooth configuration
            config = ConfigParser()
            config.optionxform = str  # Disable conversion to lowercase
            config.read("/lib/systemd/system/bluetooth.service")
            if not config["Service"]["ExecStart"].endswith(
                    "--compat --plugin=did"):
                # Update config file /lib/systemd/system/bluetooth.service to
                # start in compatible mode (compatibility with sdptool),
                # and enable only device identification service.
                config["Service"]["ExecStart"] = \
                    "/usr/libexec/bluetooth/bluetoothd --compat --plugin=did"
                with open("/lib/systemd/system/bluetooth.service", "w")\
                        as file:
                    config.write(file, space_around_delimiters=False)
                # Restart bluetoth service
                command = ["sudo", "systemctl", "daemon-reload"]
                subprocess.run(command)
                command = ["sudo", "service", "bluetooth", "restart"]
                subprocess.run(command)
            
            # Disable audio services with sdptool -> to avoid being detected
            # as an audio device by windows
            command = ["sudo", "sdptool", "del", "0x10004"]
            subprocess.run(command)
            command = ["sudo", "sdptool", "del", "0x10005"]
            subprocess.run(command)
        
        # Start bluetooth
        command = ["sudo", "btmgmt", "power", "on"]
        subprocess.run(command)
        # Set device class major=Toy minor=Robot
        # cf. https://www.bluetooth.com/specifications/assigned-numbers/
        command = ["sudo", "btmgmt", "class", "08", "04"]
        subprocess.run(command)
        # Set appearance to "generic computer"
        command = ["sudo", "btmgmt", "appearance", "80"]
        subprocess.run(command)
        # Set connectable on
        command = ["sudo", "btmgmt", "connectable", "on"]
        subprocess.run(command)
        # Set discoverable on. Must be connectable!
        command = ["sudo", "btmgmt", "discov", "yes"]
        subprocess.run(command)
    except subprocess.CalledProcessError as err:
        logger.debug("The command '" + " ".join(command)
                     + "' returned the following error message:")
        logger.debug(err.output.decode())


if old_version:
    bluetoothctl = None
else:
    # Spawn a bluetoothctl process at top level because it must keep running
    # until a connection is established
    bluetoothctl = pexpect.spawn("sudo bluetoothctl", encoding="utf-8")
    bluetoothctl.logfile = sys.stdout
    bluetoothctl.setecho(False)
    bluetoothctl.expect("Agent registered")


def init_bluetooth():
    wait_for_bt_init()
    set_bluetooth_config()
    try:
        # Setting Secure Simple Pairing mode on
        command = ["sudo", "btmgmt", "ssp", "on"]
        subprocess.run(command)
        if old_version:
            # No need for a Bluetooth agent on BlueZ version <= 5.47!
            pass
        else:
            # Starting a bluetooth agent using bluetoothctl
            # Use expect command on strings actually returned by bluetoothctl,
            # but not on the [bluetooth]# prompt, which seems unreliable.
            bluetoothctl.sendline("agent off")
            bluetoothctl.expect("Agent unregistered")

            bluetoothctl.sendline("agent NoInputNoOutput")
            bluetoothctl.expect("Agent registered")

            bluetoothctl.sendline("default-agent")
            bluetoothctl.expect("Default agent request successful")
    except Exception as err:
        print("Exception encountered in init_bluetooth:", err)
    # Setting up the new name from the ssid
    init_b_name()
    logger.info('Bluetooth has been updated')


# I do not know how this works
# I would use: from getmac import get_mac_address
# But it would require installing package get-mac
def get_b_mac():
    command = ["cat", "/sys/kernel/debug/bluetooth/hci0/identity"]
    output = subprocess.check_output(
        command, stderr=subprocess.STDOUT).decode()[0:17]
    return output


if __name__ == '__main__':
    init_bluetooth()
